"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import fnmatch
import json
import logging
import multiprocessing
import os
import re
import subprocess
import warnings
import zipfile
from io import BytesIO
from urllib.request import Request, urlopen

import jsonpatch

import cfnlint
import cfnlint.data.AdditionalSpecs
from cfnlint.helpers import SPEC_REGIONS, get_url_content, url_has_newer_version

LOGGER = logging.getLogger(__name__)

REGISTRY_SCHEMA_ZIP = (
    "https://schema.cloudformation.us-east-1.amazonaws.com/CloudformationSchema.zip"
)


def update_resource_specs(force: bool = False):
    # Pool() uses cpu count if no number of processors is specified
    # Pool() only implements the Context Manager protocol from Python3.3 onwards,
    # so it will fail Python2.7 style linting, as well as throw AttributeError
    schema_cache = get_schema_value_types()

    # pre-work us-east-1 for comparison later
    update_resource_spec("us-east-1", SPEC_REGIONS["us-east-1"], schema_cache, force)
    try:
        # pylint: disable=not-context-manager
        with multiprocessing.Pool() as pool:
            # Patch from registry schema
            pool_tuple = [
                (k, v, schema_cache, force)
                for k, v in SPEC_REGIONS.items()
                if k != "us-east-1"
            ]
            pool.starmap(update_resource_spec, pool_tuple)
    except AttributeError:
        # Do it the long, slow way
        for region, url in SPEC_REGIONS.items():
            if region != "us-east-1":
                update_resource_spec(region, url, schema_cache, force)


def update_resource_spec(region, url, schema_cache, force: bool = False):
    """Update a single resource spec"""
    filename = os.path.join(
        os.path.dirname(cfnlint.__file__), f"data/CloudSpecs/{region}.json"
    )

    multiprocessing_logger = multiprocessing.log_to_stderr()

    multiprocessing_logger.debug("Downloading template %s into %s", url, filename)

    # Check to see if we already have the latest version, and if so stop
    if not (url_has_newer_version(url) or force):
        return

    spec_content = get_url_content(url, caching=True)

    multiprocessing_logger.debug(
        "A more recent version of %s was found, and will be downloaded to %s",
        url,
        filename,
    )
    spec = json.loads(spec_content)

    # Patch the files
    spec = patch_spec(spec, "all")
    spec = patch_spec(spec, region)

    # do each patch individually so we can ignore errors
    for patch in schema_cache:
        try:
            # since there could be patched in values to ValueTypes
            # Ref/GetAtt as an example.  So we want to add new
            # ValueTypes that don't exist
            for i_patch in patch:
                path_details = i_patch.get("path").split("/")
                if path_details[1] == "ValueTypes":
                    if not spec.get("ValueTypes").get(path_details[2]):
                        spec["ValueTypes"][path_details[2]] = {}
            # do the patch
            jsonpatch.JsonPatch(patch).apply(spec, in_place=True)
        except jsonpatch.JsonPatchConflict:
            for i_patch in patch:
                path_details = i_patch.get("path").split("/")
                if path_details[1] == "ValueTypes":
                    if not spec.get("ValueTypes").get(path_details[2]):
                        try:
                            del spec["ValueTypes"][path_details[2]]
                        except:  # pylint: disable=bare-except
                            pass
            LOGGER.debug("Patch (%s) not applied in region %s", patch, region)
        except jsonpatch.JsonPointerException:
            for i_patch in patch:
                path_details = i_patch.get("path").split("/")
                if path_details[1] == "ValueTypes":
                    if not spec.get("ValueTypes").get(path_details[2]):
                        try:
                            del spec["ValueTypes"][path_details[2]]
                        except:  # pylint: disable=bare-except
                            pass
            # Debug as the parent element isn't supported in the region
            LOGGER.debug(
                "Parent element not found for patch (%s) in region %s", patch, region
            )

    botocore_cache = {}

    def search_and_replace_botocore_types(obj):
        if isinstance(obj, dict):
            new_obj = {}
            for key, value in obj.items():
                if key == "botocore":
                    service_and_type = value.split("/")
                    service = "/".join(service_and_type[:-1])
                    botocore_type = service_and_type[-1]
                    if service not in botocore_cache:
                        botocore_cache[service] = json.loads(
                            get_url_content(
                                "https://raw.githubusercontent.com/boto/botocore/master/botocore/data/"
                                + service
                                + "/service-2.json"
                            )
                        )
                    new_obj["AllowedValues"] = sorted(
                        botocore_cache[service]["shapes"][botocore_type]["enum"]
                    )
                else:
                    new_obj[key] = search_and_replace_botocore_types(value)
            return new_obj

        if isinstance(obj, list):
            new_list = []
            for item in obj:
                new_list.append(search_and_replace_botocore_types(item))
            return new_list

        return obj

    spec = search_and_replace_botocore_types(spec)

    if region != "us-east-1":
        base_specs = cfnlint.helpers.load_resource(
            "cfnlint.data.CloudSpecs", "us-east-1.json"
        )
        for section, section_values in spec.items():
            if section in ["ResourceTypes", "PropertyTypes", "ValueTypes"]:
                for key, value in section_values.items():
                    base_value = base_specs.get(section, {}).get(key, {})
                    if json.dumps(value, sort_keys=True) == json.dumps(
                        base_value, sort_keys=True
                    ):
                        spec[section][key] = "CACHED"

    with open(filename, "w", encoding="utf-8") as f:
        json.dump(spec, f, indent=1, sort_keys=True, separators=(",", ": "))


def update_documentation(rules):
    # Update the overview of all rules in the linter
    filename = "docs/rules.md"

    # Sort rules by the Rule ID
    sorted_rules = sorted(rules, key=lambda obj: obj.id)

    data = []

    # Read current file up to the Rules part, everything up to that point is
    # static documentation.
    with open(filename, "r", encoding="utf-8") as original_file:
        line = original_file.readline()
        while line:
            data.append(line)

            if line == "## Rules\n":
                break

            line = original_file.readline()

    # Rebuild the file content
    with open(filename, "w", encoding="utf-8") as new_file:
        # Rewrite the static documentation
        for line in data:
            new_file.write(line)

        # Add the rules
        new_file.write(
            "(_This documentation is generated by running `cfn-lint --update-documentation`, do not alter this manually_)\n\n"
        )
        new_file.write(
            f"The following **{len(sorted_rules) + 3}** rules are applied by this linter:\n\n"
        )
        new_file.write(
            "| Rule ID  | Title | Description | Config<br />(Name:Type:Default) | Source | Tags |\n"
        )
        new_file.write(
            "| -------- | ----- | ----------- | ---------- | ------ | ---- |\n"
        )

        rule_output = (
            '| [{0}<a name="{0}"></a>]({6}) | {1} | {2} | {3} | [Source]({4}) | {5} |\n'
        )

        for rule in [
            cfnlint.rules.ParseError(),
            cfnlint.rules.TransformError(),
            cfnlint.rules.RuleError(),
        ] + sorted_rules:
            rule_source_code_file = (
                "../"
                + subprocess.check_output(
                    [
                        "git",
                        "grep",
                        "-l",
                        'id = "' + rule.id + '"',
                        "src/cfnlint/rules/",
                    ]
                )
                .decode("ascii")
                .strip()
            )
            rule_id = rule.id + "*" if rule.experimental else rule.id
            tags = ",".join(f"`{tag}`" for tag in rule.tags)
            config = "<br />".join(
                f'{key}:{values.get("type")}:{values.get("default")}'
                for key, values in rule.config_definition.items()
            )
            new_file.write(
                rule_output.format(
                    rule_id,
                    rule.shortdesc,
                    rule.description,
                    config,
                    rule.source_url,
                    tags,
                    rule_source_code_file,
                )
            )
        new_file.write("\n\\* experimental rules\n")


def patch_spec(content, region):
    """Patch the spec file"""
    LOGGER.info('Patching spec file for region "%s"', region)

    append_dir = os.path.join(
        os.path.dirname(__file__), "data", "ExtendedSpecs", region
    )
    for dirpath, _, filenames in os.walk(append_dir):
        filenames.sort()
        for filename in fnmatch.filter(filenames, "*.json"):
            file_path = os.path.basename(filename)
            module = dirpath.replace(f"{append_dir}", f"{region}").replace(
                os.path.sep, "."
            )
            LOGGER.info("Processing %s/%s", module, file_path)
            all_patches = jsonpatch.JsonPatch(
                cfnlint.helpers.load_resource(
                    f"cfnlint.data.ExtendedSpecs.{module}", file_path
                )
            )

            # Process the generic patches 1 by 1 so we can "ignore" failed ones
            for all_patch in all_patches:
                try:
                    jsonpatch.JsonPatch([all_patch]).apply(content, in_place=True)
                except jsonpatch.JsonPatchConflict:
                    LOGGER.debug(
                        "Patch (%s) not applied in region %s", all_patch, region
                    )
                except jsonpatch.JsonPointerException:
                    # Debug as the parent element isn't supported in the region
                    LOGGER.debug(
                        "Parent element not found for patch (%s) in region %s",
                        all_patch,
                        region,
                    )

    return content


def update_iam_policies():
    """update iam policies file"""

    url = "https://awspolicygen.s3.amazonaws.com/js/policies.js"

    filename = os.path.join(
        os.path.dirname(cfnlint.data.AdditionalSpecs.__file__), "Policies.json"
    )
    LOGGER.debug("Downloading policies %s into %s", url, filename)

    content = get_url_content(url)

    content = content.split("app.PolicyEditorConfig=")[1]
    content = json.loads(content)

    actions = {
        "Manage Amazon API Gateway": ["HEAD", "OPTIONS"],
        "Amazon API Gateway Management": ["HEAD", "OPTIONS"],
        "Amazon API Gateway Management V2": ["HEAD", "OPTIONS"],
        "Amazon Kinesis Video Streams": ["StartStreamEncryption"],
    }
    for k, v in actions.items():
        if content.get("serviceMap").get(k):
            content["serviceMap"][k]["Actions"].extend(v)
        else:
            LOGGER.debug('"%s" was not found in the policies file', k)

    with open(filename, "w", encoding="utf-8") as f:
        json.dump(content, f, indent=1, sort_keys=True, separators=(",", ": "))


def get_schema_value_types():
    def resolve_refs(properties, schema):
        results = {}
        name = None

        if properties.get("$ref"):
            name = properties.get("$ref").split("/")[-1]
            subname, results = resolve_refs(schema.get("definitions").get(name), schema)
            if subname:
                name = subname
            properties = schema.get("definitions").get(name)

        if properties.get("type") == "array":
            results = properties.get("items")

        if results:
            properties = results

        if results and results.get("$ref"):
            name, results = resolve_refs(results, schema)

        if not results:
            return name, properties

        return name, results

    def get_object_details(names, properties, schema):
        results = {}
        warnings.filterwarnings("error")
        for propname, propdetails in properties.items():
            subname, propdetails = resolve_refs(propdetails, schema)
            t = propdetails.get("type")
            if not t:
                continue
            if t in ["object"]:
                if subname is None:
                    subname = propname
                if propdetails.get("properties"):
                    if subname not in names:
                        results.update(
                            get_object_details(
                                names + [subname], propdetails.get("properties"), schema
                            )
                        )
                elif (
                    propdetails.get("oneOf")
                    or propdetails.get("anyOf")
                    or propdetails.get("allOf")
                ):
                    LOGGER.info(
                        "Type %s object for %s has only oneOf,anyOf, or allOf properties",
                        names[0],
                        propname,
                    )
                    continue
            elif t not in ["string", "integer", "number", "boolean"]:
                if propdetails.get("$ref"):
                    results.update(
                        get_object_details(
                            names + [propname],
                            schema.get("definitions").get(t.get("$ref").split("/")[-1]),
                            schema,
                        )
                    )
                elif isinstance(t, list):
                    LOGGER.info(
                        "Type for %s object and %s property is a list",
                        names[0],
                        propname,
                    )
                else:
                    LOGGER.info(
                        "Unable to handle %s object for %s property", names[0], propname
                    )
            elif t == "string":
                if not results.get(".".join(names + [propname])):
                    if (
                        propdetails.get("pattern")
                        or (
                            propdetails.get("minLength")
                            and propdetails.get("maxLength")
                        )
                        or propdetails.get("enum")
                    ):
                        results[".".join(names + [propname])] = {}
                if propdetails.get("pattern"):
                    p = propdetails.get("pattern")
                    if (
                        ".".join(names + [propname])
                        == "AWS::OpsWorksCM::Server.CustomPrivateKey"
                    ):
                        # one off exception to handle a weird parsing issue in python 2.7
                        continue
                    # python 3 has the ability to test isascii
                    # python 3.7 introduces is ascii so switching to encode
                    try:
                        p.encode("ascii")
                    except UnicodeEncodeError:
                        continue
                    try:
                        if "\\p{" in p:
                            continue
                        re.compile(p, re.UNICODE)
                        results[".".join(names + [propname])].update(
                            {"AllowedPatternRegex": p}
                        )
                    except:  # pylint: disable=bare-except
                        LOGGER.info(
                            "Unable to handle regex for type %s and property %s with regex %s",
                            names[0],
                            propname,
                            p,
                        )
                if propdetails.get("minLength") and propdetails.get("maxLength"):
                    results[".".join(names + [propname])].update(
                        {
                            "StringMin": propdetails.get("minLength"),
                            "StringMax": propdetails.get("maxLength"),
                        }
                    )
                if propdetails.get("enum"):
                    results[".".join(names + [propname])].update(
                        {"AllowedValues": propdetails.get("enum")}
                    )
            elif t in ["number", "integer"]:
                if not results.get(".".join(names + [propname])):
                    if propdetails.get("minimum") and propdetails.get("maximum"):
                        results[".".join(names + [propname])] = {}
                if propdetails.get("minimum") and propdetails.get("maximum"):
                    results[".".join(names + [propname])].update(
                        {
                            "NumberMin": propdetails.get("minimum"),
                            "NumberMax": propdetails.get("maximum"),
                        }
                    )

        return results

    def process_schema(schema):
        details = get_object_details(
            [schema.get("typeName")], schema.get("properties"), schema
        )

        # Remove duplicates
        vtypes = {}
        for n, v in details.items():
            if n.count(".") > 1:
                s = n.split(".")
                vtypes[s[0] + "." + ".".join(s[-2:])] = v
            else:
                vtypes[n] = v

        patches = []
        for n, v in vtypes.items():
            patch = []
            if v:
                if n.count(".") == 2:
                    r_type = "PropertyTypes"
                else:
                    r_type = "ResourceTypes"
                element = {
                    "op": "add",
                    "path": f'/{r_type}/{".".join(n.split(".")[0:-1])}/Properties/{n.split(".")[-1]}/Value',
                    "value": {
                        "ValueType": n,
                    },
                }
                patch.append(element)
                for s, vs in v.items():
                    element = {
                        "op": "add",
                        "path": f"/ValueTypes/{n}/{s}",
                        "value": vs,
                    }
                    patch.append(element)
            if patch:
                patches.append(patch)

        return patches

    req = Request(REGISTRY_SCHEMA_ZIP)

    results = []
    with urlopen(req) as res:
        with zipfile.ZipFile(BytesIO(res.read())) as z:
            for f in z.namelist():
                with z.open(f) as d:
                    if not isinstance(d, str):
                        data = d.read()
                    else:
                        data = d
                    if isinstance(data, bytes):
                        data = data.decode("utf-8")
                    schema = json.loads(data)
                    patches = process_schema(schema)
                    results.extend(patches)

    return results



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import json

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class EventsLogGroupName(CloudFormationLintRule):
    """Check if the settings of multiple subscriptions are included for one LogGroup"""

    id = "E2529"
    shortdesc = "Check for SubscriptionFilters have beyond 2 attachments to a CloudWatch Log Group"
    description = (
        "The current limit for a CloudWatch Log Group is they can have 2 subscription filters. "
        "We will look for duplicate LogGroupNames inside Subscription Filters and make sure they are within 2. "
        "This doesn't account for any other subscription filters getting set."
    )
    source_url = "https://github.com/awslabs/serverless-application-model/blob/master/versions/2016-10-31.md#user-content-cloudwatchlogs"
    tags = ["resources", "lambda"]
    limit = 2

    def check_events_subscription_duplicated(self, cfn):
        """Check if Lambda Events Subscription is duplicated"""
        matches = []
        message = f"You can only have {self.limit} Subscription Filters per CloudWatch Log Group"

        log_group_paths = self.__get_log_group_name_list(cfn)
        for _, c in log_group_paths.items():
            if len(c) > self.limit:
                matches.append(RuleMatch(["Resources", c[2]], message))

        return matches

    def __get_log_group_name_list(self, cfn):
        log_group_paths = {}
        for value in cfn.get_resources("AWS::Logs::SubscriptionFilter").items():
            prop = value[1].get("Properties")
            log_group_name = json.dumps(prop.get("LogGroupName"))

            if log_group_name not in log_group_paths:
                log_group_paths[log_group_name] = []

            log_group_paths[log_group_name].append(value[0])
        return log_group_paths

    def match(self, cfn):
        """Check if Lambda Events Subscription is duplicated"""
        matches = []
        matches.extend(self.check_events_subscription_duplicated(cfn))
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.helpers import LIMITS
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class LimitDescription(CloudFormationLintRule):
    """Check maximum Output description size limit"""

    id = "I6012"
    shortdesc = "Output description limit"
    description = "Check the size of Output description in the template is approaching the upper limit"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/outputs-section-structure.html"
    tags = ["outputs", "limits"]

    def match(self, cfn):
        matches = []
        for output_name, output_value in cfn.template.get("Outputs", {}).items():
            description = output_value.get("Description")
            if description:
                path = ["Outputs", output_name, "Description"]
                if (
                    LIMITS["threshold"] * LIMITS["Outputs"]["description"]
                    < len(description)
                    <= LIMITS["Outputs"]["description"]
                ):
                    message = "The length of output description ({0}) is approaching the limit ({1})"
                    matches.append(
                        RuleMatch(
                            path,
                            message.format(
                                len(description), LIMITS["Outputs"]["description"]
                            ),
                        )
                    )
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import cfnlint.helpers
from cfnlint.data import AdditionalSpecs
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Exclusive(CloudFormationLintRule):
    """Check Properties Resource Configuration"""

    id = "E2520"
    shortdesc = "Check Properties that are mutually exclusive"
    description = (
        "Making sure CloudFormation properties that are exclusive are not defined"
    )
    source_url = "https://github.com/aws-cloudformation/cfn-python-lint"
    tags = ["resources"]

    def __init__(self):
        """Init"""
        super().__init__()
        exclusivespec = cfnlint.helpers.load_resource(AdditionalSpecs, "Exclusive.json")
        self.resource_types_specs = exclusivespec["ResourceTypes"]
        self.property_types_specs = exclusivespec["PropertyTypes"]
        for resource_type_spec in self.resource_types_specs:
            self.resource_property_types.append(resource_type_spec)
        for property_type_spec in self.property_types_specs:
            self.resource_sub_property_types.append(property_type_spec)

    def check(self, properties, exclusions, path, cfn):
        """Check itself"""
        matches = []

        property_sets = cfn.get_object_without_conditions(properties)
        for property_set in property_sets:
            obj = property_set["Object"].clean()
            for prop in obj:
                if prop in exclusions:
                    for excl_property in exclusions[prop]:
                        if excl_property in obj:
                            if property_set["Scenario"] is None:
                                message = (
                                    "Property {0} should NOT exist with {1} for {2}"
                                )
                                matches.append(
                                    RuleMatch(
                                        path + [prop],
                                        message.format(
                                            excl_property,
                                            prop,
                                            "/".join(map(str, path)),
                                        ),
                                    )
                                )
                            else:
                                scenario_text = " and ".join(
                                    [
                                        f'when condition "{k}" is {v}'
                                        for (k, v) in property_set["Scenario"].items()
                                    ]
                                )
                                message = (
                                    "Property {0} should NOT exist with {1} {2} for {3}"
                                )
                                matches.append(
                                    RuleMatch(
                                        path + [prop],
                                        message.format(
                                            excl_property,
                                            prop,
                                            scenario_text,
                                            "/".join(map(str, path)),
                                        ),
                                    )
                                )

        return matches

    def match_resource_sub_properties(self, properties, property_type, path, cfn):
        """Match for sub properties"""
        matches = []

        exclusions = self.property_types_specs.get(property_type, {})
        matches.extend(self.check(properties, exclusions, path, cfn))

        return matches

    def match_resource_properties(self, properties, resource_type, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        exclusions = self.resource_types_specs.get(resource_type, {})
        matches.extend(self.check(properties, exclusions, path, cfn))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Used(CloudFormationLintRule):
    """Check if Mappings are used anywhere in the template"""

    id = "W7001"
    shortdesc = "Check if Mappings are Used"
    description = "Making sure the mappings defined are used"
    source_url = "https://github.com/aws-cloudformation/cfn-python-lint"
    tags = ["mappings"]

    def match(self, cfn):
        matches = []
        findinmap_mappings = []

        mappings = cfn.template.get("Mappings", {})

        if mappings:
            # Get all "FindInMaps" that reference a Mapping
            maptrees = cfn.transform_pre["Fn::FindInMap"]
            for maptree in maptrees:
                if isinstance(maptree[-1], list):
                    map_name = maptree[-1][0]
                    if isinstance(map_name, dict):
                        self.logger.debug(
                            "Mapping Name has a function that can have too many variations. "
                            "Disabling check %s",
                            self.id,
                        )
                        return matches

                    findinmap_mappings.append(maptree[-1][0])
                else:
                    findinmap_mappings.append(maptree[-1])

            # Check if the mappings are used
            for mapname, _ in mappings.items():
                if mapname not in findinmap_mappings:
                    message = "Mapping '{0}' is defined but not used"
                    matches.append(
                        RuleMatch(["Mappings", mapname], message.format(mapname))
                    )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.helpers import LIMITS
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class LimitDescription(CloudFormationLintRule):
    """Check Template Description Size"""

    id = "I1003"
    shortdesc = "Template description limit"
    description = (
        "Check if the size of the template description is approaching the upper limit"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/cloudformation-limits.html"
    tags = ["description", "limits"]

    def match(self, cfn):
        matches = []
        description = cfn.template.get("Description", "")
        if (
            LIMITS["threshold"] * LIMITS["template"]["description"]
            < len(description)
            <= LIMITS["template"]["description"]
        ):
            message = "The template description ({0} bytes) is approaching the limit ({1} bytes)"
            matches.append(
                RuleMatch(
                    ["Description"],
                    message.format(len(description), LIMITS["template"]["description"]),
                )
            )
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.helpers import LIMITS
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class LimitAttributes(CloudFormationLintRule):
    """Check if maximum Mapping attribute limit is exceeded"""

    id = "E7012"
    shortdesc = "Mapping attribute limit not exceeded"
    description = "Check if the amount of Mapping attributes in the template is less than the upper limit"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/cloudformation-limits.html"
    tags = ["mappings", "limits"]

    def match(self, cfn):
        matches = []
        for mapping_name, mapping in cfn.template.get("Mappings", {}).items():
            for mapping_attribute_name, mapping_attribute in mapping.items():
                path = ["Mappings", mapping_name, mapping_attribute_name]
                if len(mapping_attribute) > LIMITS["Mappings"]["attributes"]:
                    message = (
                        "The amount of mapping attributes ({0}) exceeds the limit ({1})"
                    )
                    matches.append(
                        RuleMatch(
                            path,
                            message.format(
                                len(mapping_attribute), LIMITS["Mappings"]["attributes"]
                            ),
                        )
                    )
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import re

from cfnlint.helpers import FUNCTIONS
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Aliases(CloudFormationLintRule):
    """Check if CloudFront Aliases are valid domain names"""

    id = "E3013"
    shortdesc = "CloudFront Aliases"
    description = "CloudFront aliases should contain valid domain names"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-cloudfront-distribution-distributionconfig.html#cfn-cloudfront-distribution-distributionconfig-aliases"
    tags = ["properties", "cloudfront"]

    def match(self, cfn):
        """Check cloudfront Resource Parameters"""

        matches = []

        valid_domain = re.compile(
            r"^(?:[a-z0-9\*](?:[a-z0-9-]{0,61}[a-z0-9])?\.)+[a-z0-9][a-z0-9-]{0,61}[a-z0-9]$"
        )

        results = cfn.get_resource_properties(
            ["AWS::CloudFront::Distribution", "DistributionConfig"]
        )
        for result in results:
            aliases = result["Value"].get("Aliases")
            if aliases:
                for alias in aliases:
                    if isinstance(alias, str) and alias not in FUNCTIONS:
                        wildcard = alias.split(".")
                        if "*" in wildcard[1:]:
                            path = result["Path"] + ["Aliases"]
                            message = f'Invalid use of wildcards: {alias} at {"/".join(result["Path"])}'
                            matches.append(RuleMatch(path, message))
                        if not re.match(valid_domain, alias):
                            path = result["Path"] + ["Aliases"]
                            message = f'Invalid alias found: {alias} at {"/".join(result["Path"])}'
                            matches.append(RuleMatch(path, message))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.helpers import RESOURCE_SPECS
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Value(CloudFormationLintRule):
    """Check if Outputs have string values"""

    id = "E6003"
    shortdesc = "Outputs have values of strings"
    description = "Making sure the outputs have strings as values"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/outputs-section-structure.html"
    tags = ["outputs"]

    def __init__(self):
        """Init"""
        super().__init__()
        self.resourcetypes = []

    def initialize(self, cfn):
        resourcespecs = RESOURCE_SPECS[cfn.regions[0]]
        self.resourcetypes = resourcespecs["ResourceTypes"]

    def match(self, cfn):
        matches = []

        template = cfn.template

        getatts = cfn.search_deep_keys("Fn::GetAtt")
        refs = cfn.search_deep_keys("Ref")
        # If using a getatt make sure the attribute of the resource
        # is not of Type List
        for getatt in getatts:
            if getatt[0] == "Outputs":
                if getatt[2] == "Value":
                    obj = getatt[-1]
                    if isinstance(obj, list):
                        objtype = (
                            template.get("Resources", {}).get(obj[0], {}).get("Type")
                        )
                        if objtype:
                            attribute = (
                                self.resourcetypes.get(objtype, {})
                                .get("Attributes", {})
                                .get(obj[1], {})
                                .get("Type")
                            )
                            if attribute == "List":
                                if getatt[-4] != "Fn::Join" and getatt[-3] != 1:
                                    message = "Output {0} value {1} is of type list"
                                    matches.append(
                                        RuleMatch(
                                            getatt,
                                            message.format(getatt[1], "/".join(obj)),
                                        )
                                    )

        # If using a ref for an output make sure it isn't a
        # Parameter of Type List
        for ref in refs:
            if ref[0] == "Outputs":
                if ref[2] == "Value":
                    obj = ref[-1]
                    if isinstance(obj, str):
                        param = template.get("Parameters", {}).get(obj)
                        if param:
                            paramtype = param.get("Type")
                            if paramtype:
                                if paramtype.startswith("List<"):
                                    if ref[-4] != "Fn::Join" and ref[-3] != 1:
                                        message = "Output {0} value {1} is of type list"
                                        matches.append(
                                            RuleMatch(ref, message.format(ref[1], obj))
                                        )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.helpers import RESOURCE_SPECS
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class ListSize(CloudFormationLintRule):
    """Check if List has a size within the limit"""

    id = "E3032"
    shortdesc = "Check if a list has between min and max number of values specified"
    description = "Check lists for the number of items in the list to validate they are between the minimum and maximum"
    source_url = "https://github.com/awslabs/cfn-python-lint/blob/main/docs/cfn-resource-specification.md#allowedpattern"
    tags = ["resources", "property", "list", "size"]

    def initialize(self, cfn):
        """Initialize the rule"""
        for resource_type_spec in RESOURCE_SPECS.get(cfn.regions[0]).get(
            "ResourceTypes"
        ):
            self.resource_property_types.append(resource_type_spec)
        for property_type_spec in RESOURCE_SPECS.get(cfn.regions[0]).get(
            "PropertyTypes"
        ):
            self.resource_sub_property_types.append(property_type_spec)

    def check_value(self, properties, path, property_name, cfn, value_specs):
        """Check Value"""
        matches = []

        # Get the Min and Max for a List
        list_min = value_specs.get("ListMin")
        list_max = value_specs.get("ListMax")

        if list_min is not None and list_max is not None:
            property_sets = cfn.get_object_without_conditions(properties)
            for property_set in property_sets:
                prop = property_set.get("Object").get(property_name)
                if isinstance(prop, list):
                    if not list_min <= len(prop) <= list_max:
                        if property_set["Scenario"] is None:
                            message = (
                                "{0} has to have between {1} and {2} items specified"
                            )
                            matches.append(
                                RuleMatch(
                                    path + [property_name],
                                    message.format(property_name, list_min, list_max),
                                )
                            )
                        else:
                            scenario_text = " and ".join(
                                [
                                    f'when condition "{k}" is {v}'
                                    for (k, v) in property_set["Scenario"].items()
                                ]
                            )
                            message = "{0} has to have between {1} and {2} items specified when {3}"
                            matches.append(
                                RuleMatch(
                                    path + [property_name],
                                    message.format(
                                        property_name, list_min, list_max, scenario_text
                                    ),
                                )
                            )

        return matches

    def check(self, cfn, properties, specs, path):
        """Check itself"""
        matches = []
        for p_value, p_path in properties.items_safe(path[:]):
            for prop in p_value:
                if prop in specs:
                    value = specs.get(prop).get("Value", {})
                    if value:
                        value_type = value.get("ListValueType", "")
                        property_type = specs.get(prop).get("Type")
                        if property_type == "List":
                            matches.extend(
                                self.check_value(
                                    p_value,
                                    p_path,
                                    prop,
                                    cfn,
                                    RESOURCE_SPECS.get(cfn.regions[0])
                                    .get("ValueTypes")
                                    .get(value_type, {}),
                                )
                            )
        return matches

    def match_resource_sub_properties(self, properties, property_type, path, cfn):
        """Match for sub properties"""
        matches = []

        specs = (
            RESOURCE_SPECS.get(cfn.regions[0])
            .get("PropertyTypes")
            .get(property_type, {})
            .get("Properties", {})
        )
        matches.extend(self.check(cfn, properties, specs, path))

        return matches

    def match_resource_properties(self, properties, resource_type, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        specs = (
            RESOURCE_SPECS.get(cfn.regions[0])
            .get("ResourceTypes")
            .get(resource_type, {})
            .get("Properties", {})
        )
        matches.extend(self.check(cfn, properties, specs, path))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule
from cfnlint.rules.common import alphanumeric_name


class Name(CloudFormationLintRule):
    """Check if Mappings are named correctly"""

    id = "E7002"
    shortdesc = "Mappings have appropriate names"
    description = "Check if Mappings are properly named (A-Za-z0-9)"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/mappings-section-structure.html"
    tags = ["mappings"]

    def match(self, cfn):
        return alphanumeric_name(cfn, "Mappings")



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import os

from cfnlint.decode import decode
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class NestedStackParameters(CloudFormationLintRule):
    """Check that nested stack parameters are specified as needed"""

    id = "E3043"
    shortdesc = "Validate parameters for in a nested stack"
    description = (
        "Evalute if parameters for a nested stack are specified and "
        "if parameters are specified for a nested stack that aren't required."
    )
    source_url = "https://github.com/awslabs/cfn-python-lint"
    tags = ["resources", "cloudformation"]

    def __init__(self):
        """Init"""
        super().__init__()
        self.resource_property_types.append("AWS::CloudFormation::Stack")

    def __get_template_parameters(self, filename):
        try:
            (tmp, matches) = decode(filename)
        except:  # pylint: disable=bare-except
            return None
        if matches:
            return None

        return tmp.get("Parameters", {})

    def __compare_objects(self, template_parameters, nested_parameters, scenario, path):
        matches = []
        for key in set(template_parameters.keys()) - set(nested_parameters.keys()):
            if scenario is None:
                message = 'Specified parameter "{0}" doesn\'t exist in nested stack template at {1}'
                matches.append(
                    RuleMatch(
                        path + [key],
                        message.format(key, "/".join(map(str, path + [key]))),
                    )
                )
            else:
                message = 'Specified parameter "{0}" doesn\'t exist in nested stack template {1}'
                scenario_text = " and ".join(
                    [f'when condition "{k}" is {v}' for (k, v) in scenario.items()]
                )
                matches.append(RuleMatch(path, message.format(key, scenario_text)))
        for key in set(nested_parameters.keys()) - set(template_parameters.keys()):
            if nested_parameters.get(key).get("Default") is None:
                if scenario is None:
                    message = (
                        'Nested stack template parameter "{0}" is not specified at {1}'
                    )
                    matches.append(
                        RuleMatch(path, message.format(key, "/".join(map(str, path))))
                    )
                else:
                    message = (
                        'Nested stack template parameter "{0}" is not specified {1}'
                    )
                    scenario_text = " and ".join(
                        [f'when condition "{k}" is {v}' for (k, v) in scenario.items()]
                    )
                    matches.append(RuleMatch(path, message.format(key, scenario_text)))

        return matches

    def match_resource_properties(self, properties, _, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        # when template is passed via cat (or equivalent filename is none)
        if cfn.filename:
            base_dir = os.path.dirname(os.path.abspath(cfn.filename))

            parameter_groups = cfn.get_object_without_conditions(
                obj=properties, property_names=["TemplateURL", "Parameters"]
            )
            for parameter_group in parameter_groups:
                obj = parameter_group.get("Object")
                template_url = obj.get("TemplateURL")
                if isinstance(template_url, str):
                    if not (
                        template_url.startswith("http://")
                        or template_url.startswith("https://")
                        or template_url.startswith("s3://")
                    ):
                        template_path = os.path.normpath(
                            os.path.join(base_dir, template_url)
                        )
                        nested_parameters = self.__get_template_parameters(
                            template_path
                        )
                        template_parameters = obj.get("Parameters")
                        if isinstance(nested_parameters, dict) and isinstance(
                            template_parameters, dict
                        ):
                            matches.extend(
                                self.__compare_objects(
                                    template_parameters=template_parameters,
                                    nested_parameters=nested_parameters,
                                    path=path + ["Parameters"],
                                    scenario=parameter_group.get("Scenario"),
                                )
                            )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import re

from cfnlint.helpers import RESOURCE_SPECS
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class AllowedPattern(CloudFormationLintRule):
    """Check if properties have a valid value"""

    id = "E3031"
    shortdesc = "Check if property values adhere to a specific pattern"
    description = "Check if properties have a valid value in case of a pattern (Regular Expression)"
    source_url = "https://github.com/awslabs/cfn-python-lint/blob/main/docs/cfn-resource-specification.md#allowedpattern"
    tags = ["resources", "property", "allowed pattern", "regex"]

    def __init__(self):
        """Init"""
        super().__init__()
        self.config_definition = {
            "exceptions": {
                "default": [],
                "type": "list",
                "itemtype": "string",
            }
        }
        self.configure()

    def initialize(self, cfn):
        """Initialize the rule"""
        for resource_type_spec in RESOURCE_SPECS.get(cfn.regions[0]).get(
            "ResourceTypes"
        ):
            self.resource_property_types.append(resource_type_spec)
        for property_type_spec in RESOURCE_SPECS.get(cfn.regions[0]).get(
            "PropertyTypes"
        ):
            self.resource_sub_property_types.append(property_type_spec)

    def check_value(self, value, path, property_name, **kwargs):
        """Check Value"""
        matches = []

        # Get the Allowed Pattern Regex
        value_pattern_regex = kwargs.get("value_specs", {}).get(
            "AllowedPatternRegex", {}
        )
        # Get the "Human Readable" version for the error message. Optional, if not specified,
        # the RegEx itself is used.
        value_pattern = kwargs.get("value_specs", {}).get(
            "AllowedPattern", value_pattern_regex
        )

        if isinstance(value, (int, float)):
            value = str(value)

        if isinstance(value, str):
            if value_pattern_regex:
                regex = re.compile(value_pattern_regex, re.ASCII)

                # Ignore values with dynamic references. Simple check to prevent false-positives
                # See: https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/dynamic-references.html
                if "{{resolve:" not in value:
                    if not regex.match(value):
                        for exception in self.config.get("exceptions"):
                            exception_regex = re.compile(exception)
                            if exception_regex.match(value):
                                return matches
                        full_path = "/".join(str(x) for x in path)

                        message = "{} contains invalid characters (Pattern: {}) at {}"
                        matches.append(
                            RuleMatch(
                                path,
                                message.format(property_name, value_pattern, full_path),
                            )
                        )

        return matches

    def check(self, cfn, properties, value_specs, property_specs, path):
        """Check itself"""
        matches = []
        for p_value, p_path in properties.items_safe(path[:]):
            for prop in p_value:
                if prop in value_specs:
                    value = value_specs.get(prop).get("Value", {})
                    if value:
                        value_type = value.get("ValueType", "")
                        property_type = (
                            property_specs.get("Properties").get(prop).get("Type")
                        )
                        value_specs = (
                            RESOURCE_SPECS.get(cfn.regions[0])
                            .get("ValueTypes")
                            .get(value_type, {})
                        )
                        if value_specs == "CACHED":
                            value_specs = (
                                RESOURCE_SPECS.get("us-east-1")
                                .get("ValueTypes")
                                .get(value_type, {})
                            )
                        matches.extend(
                            cfn.check_value(
                                p_value,
                                prop,
                                p_path,
                                check_value=self.check_value,
                                value_specs=value_specs,
                                cfn=cfn,
                                property_type=property_type,
                                property_name=prop,
                            )
                        )
        return matches

    def match_resource_sub_properties(self, properties, property_type, path, cfn):
        """Match for sub properties"""
        matches = []

        specs = (
            RESOURCE_SPECS.get(cfn.regions[0])
            .get("PropertyTypes")
            .get(property_type, {})
            .get("Properties", {})
        )
        property_specs = (
            RESOURCE_SPECS.get(cfn.regions[0]).get("PropertyTypes").get(property_type)
        )
        matches.extend(self.check(cfn, properties, specs, property_specs, path))

        return matches

    def match_resource_properties(self, properties, resource_type, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        specs = (
            RESOURCE_SPECS.get(cfn.regions[0])
            .get("ResourceTypes")
            .get(resource_type, {})
            .get("Properties", {})
        )
        resource_specs = (
            RESOURCE_SPECS.get(cfn.regions[0]).get("ResourceTypes").get(resource_type)
        )
        matches.extend(self.check(cfn, properties, specs, resource_specs, path))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.decode.node import sub_node
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class SubParametersUsed(CloudFormationLintRule):
    """Check if Sub Parameters are used"""

    id = "W1019"
    shortdesc = "Sub validation of parameters"
    description = "Validate that Fn::Sub Parameters are used"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-sub.html"
    tags = ["functions", "sub"]

    def match(self, cfn):
        matches = []

        sub_objs = cfn.search_deep_class(sub_node)

        for sub_obj_tuple in sub_objs:
            path, sub_obj = sub_obj_tuple
            result_path = path[:] + ["Fn::Sub", 1]
            sub_string_vars = sub_obj.get_string_vars()
            sub_vars = sub_obj.get_defined_vars()
            for sub_var in sub_vars:
                if sub_var not in sub_string_vars:
                    message = "Parameter {0} not used in Fn::Sub at {1}"
                    matches.append(
                        RuleMatch(
                            result_path + [sub_var],
                            message.format(
                                sub_var, "/".join(map(str, result_path + [sub_var]))
                            ),
                        )
                    )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class RefInCondition(CloudFormationLintRule):
    """Check if Ref value is a string"""

    id = "E1026"
    shortdesc = "Cannot reference resources in the Conditions block of the template"
    description = "Check that any Refs in the Conditions block uses no resources"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-conditions.html#w2ab2c21c28c21c45"
    tags = ["conditions", "functions", "ref"]

    def match(self, cfn):
        matches = []

        ref_objs = cfn.search_deep_keys("Ref")
        resource_names = cfn.get_resource_names()

        for ref_obj in ref_objs:
            if ref_obj[0] == "Conditions":
                value = ref_obj[-1]
                if isinstance(value, (str, int)):
                    if value in resource_names:
                        message = "Cannot reference resource {0} in the Conditions block of the template at {1}"
                        matches.append(
                            RuleMatch(
                                ref_obj[:-1],
                                message.format(value, "/".join(map(str, ref_obj[:-1]))),
                            )
                        )

        return matches



"""
Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import logging
import re
from copy import copy, deepcopy
from typing import Union

import cfnlint.conditions
import cfnlint.helpers
from cfnlint.graph import Graph

LOGGER = logging.getLogger(__name__)


class Template:  # pylint: disable=R0904,too-many-lines,too-many-instance-attributes
    """Class for a CloudFormation template"""

    # pylint: disable=dangerous-default-value
    def __init__(self, filename, template, regions=["us-east-1"]):
        self.filename = filename
        self.template = template
        self.regions = regions
        self.sections = [
            "AWSTemplateFormatVersion",
            "Description",
            "Metadata",
            "Parameters",
            "Mappings",
            "Conditions",
            "Transform",
            "Hooks",
            "Resources",
            "Outputs",
            "Rules",
        ]
        self.transform_pre = {}
        self.transform_pre["Globals"] = {}
        self.transform_pre["Ref"] = self.search_deep_keys("Ref")
        self.transform_pre["Fn::Sub"] = self.search_deep_keys("Fn::Sub")
        self.transform_pre["Fn::FindInMap"] = self.search_deep_keys("Fn::FindInMap")
        self.transform_pre["Transform"] = self.template.get("Transform", [])
        self.conditions = cfnlint.conditions.Conditions(self)
        self.__cache_search_deep_class = {}
        self.graph: Union[Graph, None] = None
        try:
            self.graph = Graph(self)
        except KeyError as err:
            LOGGER.debug(
                "Encountered KeyError error while building graph. Ignored as this "
                "should be caught by other rules and is more than likely a template "
                "formatting error: %s",
                err,
            )
        except Exception as err:  # pylint: disable=broad-except
            LOGGER.info("Encountered unknown error while building graph: %s", err)

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, deepcopy(v, memo))
        return result

    def build_graph(self):
        """Generates a DOT representation of the template"""
        path = self.filename + ".dot"
        try:
            self.graph.to_dot(path)
            LOGGER.info("DOT representation of the graph written to %s", path)
        except ImportError:
            LOGGER.error(
                "Could not write the graph in DOT format. Please install either `pygraphviz` or `pydot` modules."
            )

    def has_language_extensions_transform(self):
        """Check if the template has language extensions transform declared"""
        LOGGER.debug(
            "Check if the template has language extensions transform declaration"
        )
        lang_extensions_transform = "AWS::LanguageExtensions"
        transform_declaration = self.transform_pre["Transform"]
        transform_type = (
            transform_declaration
            if isinstance(transform_declaration, list)
            else [transform_declaration]
        )
        return bool(lang_extensions_transform in transform_type)

    def get_resources(self, resource_type=[]):
        """
        Get Resources
        Filter on type when specified
        """
        LOGGER.debug("Get resources from template...")
        resources = self.template.get("Resources", {})
        if not isinstance(resources, dict):
            return {}
        if isinstance(resource_type, str):
            resource_type = [resource_type]

        results = {}
        for k, v in resources.items():
            if isinstance(v, dict):
                if (v.get("Type", None) in resource_type) or (
                    not resource_type and v.get("Type") is not None
                ):
                    results[k] = v

        return results

    def get_parameters(self):
        LOGGER.debug("Get parameters from template...")
        parameters = self.template.get("Parameters", {})
        if not parameters:
            return {}

        return parameters

    def get_parameters_valid(self):
        LOGGER.debug("Get parameters from template...")
        result = {}
        if isinstance(self.template.get("Parameters"), dict):
            parameters = self.template.get("Parameters")
            for parameter_name, parameter_value in parameters.items():
                if isinstance(parameter_value, dict):
                    if isinstance(parameter_value.get("Type"), str):
                        result[parameter_name] = parameter_value

        return result

    def get_outputs_valid(self):
        LOGGER.debug("Get outputs from template...")
        result = {}
        if isinstance(self.template.get("Outputs"), dict):
            parameters = self.template.get("Outputs")
            for parameter_name, parameter_value in parameters.items():
                if isinstance(parameter_value, dict):
                    if isinstance(parameter_value.get("Value"), (str, dict)):
                        result[parameter_name] = parameter_value

        return result

    def get_modules(self):
        """Get Modules"""
        LOGGER.debug("Get modules from template...")
        resources = self.template.get("Resources", {})
        if not resources:
            return {}

        results = {}
        for k, v in resources.items():
            if isinstance(v, dict):
                if v.get("Type") is not None and str(v.get("Type")).endswith(
                    "::MODULE"
                ):
                    results[k] = v

        return results

    def get_mappings(self):
        LOGGER.debug("Get mapping from template...")
        mappings = self.template.get("Mappings", {})
        if not mappings:
            return {}

        return mappings

    def get_resource_names(self):
        LOGGER.debug("Get the names of all resources from template...")
        results = []
        resources = self.template.get("Resources", {})
        if isinstance(resources, dict):
            for resourcename, _ in resources.items():
                results.append(resourcename)

        return results

    def get_parameter_names(self):
        LOGGER.debug("Get names of all parameters from template...")
        results = []
        parameters = self.template.get("Parameters", {})
        if isinstance(parameters, dict):
            for parametername, _ in parameters.items():
                results.append(parametername)

        return results

    def get_valid_refs(self):
        results = cfnlint.helpers.RegexDict()
        parameters = self.template.get("Parameters", {})
        if parameters:
            for name, value in parameters.items():
                if isinstance(value, dict):
                    if "Type" in value:
                        element = {}
                        element["Type"] = value["Type"]
                        element["From"] = "Parameters"
                        results[name] = element
        resources = self.template.get("Resources", {})
        if resources:
            for name, value in resources.items():
                resource_type = value.get("Type", "")
                if resource_type.endswith("::MODULE"):
                    element = {}
                    element["Type"] = "MODULE"
                    element["From"] = "Resources"
                    results[f"{name}.*"] = element
                elif resource_type:
                    element = {}
                    element["Type"] = resource_type
                    element["From"] = "Resources"
                    results[name] = element

        for pseudoparam in cfnlint.PSEUDOPARAMS:
            element = {}
            element["Type"] = "Pseudo"
            element["From"] = "Pseduo"
            results[pseudoparam] = element
        return results

    def get_valid_getatts(self):
        resourcetypes = cfnlint.helpers.RESOURCE_SPECS["us-east-1"].get("ResourceTypes")
        propertytypes = cfnlint.helpers.RESOURCE_SPECS["us-east-1"].get("PropertyTypes")
        results = {}
        resources = self.template.get("Resources", {})

        astrik_string_types = ("AWS::CloudFormation::Stack",)
        astrik_unknown_types = (
            "Custom::",
            "AWS::Serverless::",
            "AWS::CloudFormation::CustomResource",
        )

        def build_output_string(resource_type, property_name):
            prop = propertytypes.get(f"{resource_type}.{property_name}")
            if prop is None:
                yield None, None
            else:
                for k, v in prop.get("Properties", {}).items():
                    t = v.get("Type")
                    if t:
                        for item in build_output_string(resource_type, v):
                            yield f"{k}.{item[0]}", item[1]
                    else:
                        yield k, v.get("PrimitiveType")

        for name, value in resources.items():
            if "Type" in value:
                valtype = value["Type"]
                if isinstance(valtype, str):
                    if valtype.startswith(astrik_string_types):
                        LOGGER.debug(
                            "Cant build an appropriate getatt list from %s", valtype
                        )
                        results[name] = {"*": {"PrimitiveItemType": "String"}}
                    elif valtype.startswith(astrik_unknown_types) or valtype.endswith(
                        "::MODULE"
                    ):
                        LOGGER.debug(
                            "Cant build an appropriate getatt list from %s", valtype
                        )
                        results[name] = {"*": {}}
                    else:
                        if value["Type"] in resourcetypes:
                            if "Attributes" in resourcetypes[valtype]:
                                results[name] = {}
                                for attname, attvalue in resourcetypes[valtype][
                                    "Attributes"
                                ].items():
                                    if "Type" in attvalue:
                                        if attvalue.get("Type") in ["List", "Map"]:
                                            element = {}
                                            element.update(attvalue)
                                            results[name][attname] = element
                                        else:
                                            for item in build_output_string(
                                                value["Type"], attname
                                            ):
                                                if item[0] is None:
                                                    continue
                                                element = {"PrimitiveType": item[1]}
                                                results[name][
                                                    f"{attname}.{item[0]}"
                                                ] = element
                                    else:
                                        element = {}
                                        element.update(attvalue)
                                        results[name][attname] = element

        return results

    def get_directives(self):
        results = {}
        resources = self.get_resources()
        if resources:
            for resource_name, resource_values in resources.items():
                if isinstance(resource_values, dict):
                    ignore_rule_ids = (
                        resource_values.get("Metadata", {})
                        .get("cfn-lint", {})
                        .get("config", {})
                        .get("ignore_checks", [])
                    )
                    for ignore_rule_id in ignore_rule_ids:
                        if ignore_rule_id not in results:
                            results[ignore_rule_id] = []
                        value_location = self._loc(resource_values)
                        name_location = self._loc(resource_name)
                        results[ignore_rule_id].append(
                            {
                                "start": (name_location[0] + 1, name_location[1] + 1),
                                "end": (value_location[2] + 1, value_location[3] + 1),
                            }
                        )
        return results

    # pylint: disable=too-many-locals
    def _get_sub_resource_properties(self, keys, properties, path):
        """Used for recursive handling of properties in the keys"""
        LOGGER.debug("Get Sub Resource Properties from %s", keys)
        if not keys:
            result = {}
            result["Path"] = path
            result["Value"] = properties
            return [result]
        if isinstance(properties, dict):
            key = keys.pop(0)
            for key_name, key_value in properties.items():
                if key_name == key:
                    results = self._get_sub_resource_properties(
                        keys[:], key_value, path[:] + [key_name]
                    )
                    if results:
                        return results
        elif isinstance(properties, list):
            matches = []
            for index, item in enumerate(properties):
                results = None
                if isinstance(item, dict):
                    if len(item) == 1:
                        for sub_key, sub_value in item.items():
                            if sub_key in cfnlint.helpers.CONDITION_FUNCTIONS:
                                cond_values = self.get_condition_values(sub_value)
                                results = []
                                for cond_value in cond_values:
                                    result_path = (
                                        path[:] + [index, sub_key] + cond_value["Path"]
                                    )
                                    results.extend(
                                        self._get_sub_resource_properties(
                                            keys[:], cond_value["Value"], result_path
                                        )
                                    )
                            elif sub_key == "Ref":
                                if sub_value != "AWS::NoValue":
                                    results = self._get_sub_resource_properties(
                                        keys[:], sub_value, path + [sub_key]
                                    )
                            else:
                                results = self._get_sub_resource_properties(
                                    keys[:], sub_value, path + [sub_key]
                                )
                    else:
                        results = self._get_sub_resource_properties(
                            keys[:], item, path + [index]
                        )
                if isinstance(results, dict):
                    matches.append(results)
                elif isinstance(results, list):
                    matches.extend(results)
            return matches

        return []

    def get_resource_properties(self, keys):
        """Filter keys of template"""
        LOGGER.debug("Get Properties from a resource: %s", keys)
        matches = []
        resourcetype = keys.pop(0)
        for resource_name, resource_value in self.get_resources(resourcetype).items():
            path = ["Resources", resource_name, "Properties"]
            properties = resource_value.get("Properties")
            if properties:
                results = self._get_sub_resource_properties(keys[:], properties, path)
                matches.extend(results)

        return matches

    # pylint: disable=dangerous-default-value
    def _search_deep_class(self, searchClass, cfndict, path):
        """Search deep for keys and get their values"""
        keys = []
        if isinstance(cfndict, searchClass):
            keys.append((path[:], cfndict))

        if isinstance(cfndict, dict):
            for key in cfndict:
                keys.extend(
                    self._search_deep_class(searchClass, cfndict[key], path[:] + [key])
                )
        elif isinstance(cfndict, list):
            for index, item in enumerate(cfndict):
                keys.extend(
                    self._search_deep_class(searchClass, item, path[:] + [index])
                )

        return keys

    def search_deep_class(self, searchClass, includeGlobals=True):
        """
        Search for a key in all parts of the template.
        :return if searchText is "Ref", an array like ['Resources', 'myInstance', 'Properties', 'ImageId', 'Ref', 'Ec2ImageId']
        """
        results = []
        if searchClass in self.__cache_search_deep_class:
            results = self.__cache_search_deep_class[searchClass]
        else:
            results.extend(self._search_deep_class(searchClass, self.template, []))
            self.__cache_search_deep_class[searchClass] = results

        # Globals are removed during a transform.  They need to be checked manually
        if includeGlobals:
            pre_results = self._search_deep_keys(
                searchClass, self.transform_pre.get("Globals"), []
            )
            for pre_result in pre_results:
                results.append(["Globals"] + pre_result)

        return results

    # pylint: disable=dangerous-default-value
    def _search_deep_keys(self, searchText, cfndict, path):
        """Search deep for keys and get their values"""
        keys = []
        if isinstance(cfndict, dict):
            for key in cfndict:
                pathprop = path[:]
                pathprop.append(key)
                if key == searchText:
                    pathprop.append(cfndict[key])
                    keys.append(pathprop)
                    # pop the last element off for nesting of found elements for
                    # dict and list checks
                    pathprop = pathprop[:-1]
                if isinstance(cfndict[key], dict):
                    keys.extend(
                        self._search_deep_keys(searchText, cfndict[key], pathprop)
                    )
                elif isinstance(cfndict[key], list):
                    for index, item in enumerate(cfndict[key]):
                        pathproparr = pathprop[:]
                        pathproparr.append(index)
                        keys.extend(
                            self._search_deep_keys(searchText, item, pathproparr)
                        )
        elif isinstance(cfndict, list):
            for index, item in enumerate(cfndict):
                pathprop = path[:]
                pathprop.append(index)
                keys.extend(self._search_deep_keys(searchText, item, pathprop))

        return keys

    def search_deep_keys(self, searchText, includeGlobals=True):
        """
        Search for a key in all parts of the template.
        :return if searchText is "Ref", an array like ['Resources', 'myInstance', 'Properties', 'ImageId', 'Ref', 'Ec2ImageId']
        """
        LOGGER.debug("Search for key %s as far down as the template goes", searchText)
        results = []
        results.extend(self._search_deep_keys(searchText, self.template, []))
        # Globals are removed during a transform.  They need to be checked manually
        if includeGlobals:
            pre_results = self._search_deep_keys(
                searchText, self.transform_pre.get("Globals"), []
            )
            for pre_result in pre_results:
                results.append(["Globals"] + pre_result)
        return results

    def get_condition_values(self, template, path=[]):
        """Evaluates conditions and brings back the values"""
        LOGGER.debug("Get condition values...")
        matches = []
        if not isinstance(template, list):
            return matches
        if not len(template) == 3:
            return matches

        for index, item in enumerate(template[1:]):
            result = {}
            result["Path"] = path[:] + [index + 1]
            if not isinstance(item, (dict, list)):
                # Just straight values and pass them through
                result["Value"] = item
                matches.append(result)
            elif len(item) == 1:
                # Checking for conditions inside of conditions
                if isinstance(item, dict):
                    for sub_key, sub_value in item.items():
                        if sub_key in cfnlint.helpers.CONDITION_FUNCTIONS:
                            results = self.get_condition_values(
                                sub_value, result["Path"] + [sub_key]
                            )
                            if isinstance(results, list):
                                matches.extend(results)
                        elif sub_key == "Ref":
                            if sub_value != "AWS::NoValue":
                                result["Value"] = item
                                matches.append(result)
                        else:
                            # Return entire Item
                            result["Value"] = item
                            matches.append(result)
                else:
                    # Return entire Item
                    result["Value"] = item
                    matches.append(result)
            else:
                # Length longer than 1 means a list or object that should be fully returned
                result["Value"] = item
                matches.append(result)

        return matches

    def get_values(self, obj, key, path=[]):
        """
        Logic for getting the value of a key
        Returns None if the item isn't found
        Returns empty list if the item is found but Ref or GetAtt
        Returns all the values as a list if condition
        Returns the value if its just a string, int, boolean, etc.

        """
        LOGGER.debug("Get the value for key %s in %s", key, obj)
        matches = []

        if not isinstance(obj, dict):
            return None
        value = obj.get(key)
        if value is None:
            return None
        if isinstance(value, (dict)):
            if len(value) == 1:
                is_condition = False
                is_no_value = False
                for obj_key, obj_value in value.items():
                    if obj_key in cfnlint.helpers.CONDITION_FUNCTIONS:
                        is_condition = True
                        results = self.get_condition_values(
                            obj_value, path[:] + [obj_key]
                        )
                        if isinstance(results, list):
                            for result in results:
                                check_obj = obj.copy()
                                check_obj[key] = result["Value"]
                                matches.extend(
                                    self.get_values(check_obj, key, result["Path"])
                                )
                    elif obj_key == "Ref" and obj_value == "AWS::NoValue":
                        is_no_value = True
                if not is_condition and not is_no_value:
                    result = {}
                    result["Path"] = path[:]
                    result["Value"] = value
                    matches.append(result)
            else:
                result = {}
                result["Path"] = path[:]
                result["Value"] = value
                matches.append(result)
        elif isinstance(value, (list)):
            for list_index, list_value in enumerate(value):
                if isinstance(list_value, dict):
                    if len(list_value) == 1:
                        is_condition = False
                        is_no_value = False
                        for obj_key, obj_value in list_value.items():
                            if obj_key in cfnlint.helpers.CONDITION_FUNCTIONS:
                                is_condition = True
                                results = self.get_condition_values(
                                    obj_value, path[:] + [list_index, obj_key]
                                )
                                if isinstance(results, list):
                                    matches.extend(results)
                            elif obj_key == "Ref" and obj_value == "AWS::NoValue":
                                is_no_value = True
                        if not is_condition and not is_no_value:
                            result = {}
                            result["Path"] = path[:] + [list_index]
                            result["Value"] = list_value
                            matches.append(result)
                    else:
                        result = {}
                        result["Path"] = path[:] + [list_index]
                        result["Value"] = list_value
                        matches.append(result)
                else:
                    result = {}
                    result["Path"] = path[:] + [list_index]
                    result["Value"] = list_value
                    matches.append(result)
        else:
            result = {}
            result["Path"] = path[:]
            result["Value"] = value
            matches.append(result)

        return matches

    def _loc(self, obj):
        """Return location of object"""
        LOGGER.debug("Get location of object...")
        return (
            obj.start_mark.line,
            obj.start_mark.column,
            obj.end_mark.line,
            obj.end_mark.column,
        )

    def get_sub_parameters(self, sub_string):
        """Gets the parameters out of a Sub String"""
        results = []
        if not isinstance(sub_string, str):
            return results
        regex = re.compile(r"\${[^!].*?}")
        string_params = regex.findall(sub_string)

        for string_param in string_params:
            results.append(string_param[2:-1].strip())

        return results

    def get_location_yaml(self, text, path):
        """
        Get the location information
        """
        LOGGER.debug("Get location of path %s", path)
        result = None
        if not path:
            result = self._loc(text)
        elif len(path) > 1:
            try:
                result = self.get_location_yaml(text[path[0]], path[1:])
            except KeyError:
                pass
            # TypeError will help catch string indices must be integers for when
            # we parse JSON string and get a path inside that json string
            except TypeError:
                pass
            if not result:
                try:
                    for key in text:
                        if key == path[0]:
                            result = self._loc(key)
                except AttributeError as err:
                    LOGGER.debug(err)
        else:
            # If the last item of the path is an integer, and the vaue is an array,
            # Get the location of the item in the array
            if isinstance(text, list) and isinstance(path[0], int):
                try:
                    result = self._loc(text[path[0]])
                except AttributeError as err:
                    LOGGER.debug(err)
            else:
                try:
                    for key in text:
                        if key == path[0]:
                            result = self._loc(key)
                except AttributeError as err:
                    LOGGER.debug(err)

        return result

    def check_resource_property(
        self,
        resource_type,
        resource_property,
        check_value=None,
        check_ref=None,
        check_find_in_map=None,
        check_split=None,
        check_join=None,
        check_sub=None,
        **kwargs,
    ):
        """Check Resource Properties"""
        LOGGER.debug("Check property %s for %s", resource_property, resource_type)
        matches = []
        resources = self.get_resources(resource_type=resource_type)
        for resource_name, resource_object in resources.items():
            properties = resource_object.get("Properties", {})
            if properties:
                matches.extend(
                    self.check_value(
                        obj=properties,
                        key=resource_property,
                        path=["Resources", resource_name, "Properties"],
                        check_value=check_value,
                        check_ref=check_ref,
                        check_find_in_map=check_find_in_map,
                        check_split=check_split,
                        check_join=check_join,
                        check_sub=check_sub,
                        **kwargs,
                    )
                )
        return matches

    # pylint: disable=W0613,too-many-locals
    def check_value(
        self,
        obj,
        key,
        path,
        check_value=None,
        check_ref=None,
        check_get_att=None,
        check_find_in_map=None,
        check_split=None,
        check_join=None,
        check_import_value=None,
        check_sub=None,
        pass_if_null=False,
        **kwargs,
    ):
        LOGGER.debug("Check value %s for %s", key, obj)
        matches = []
        values_obj = self.get_values(obj=obj, key=key)
        new_path = path[:] + [key]
        if values_obj is None and pass_if_null:
            if check_value:
                matches.extend(
                    check_value(value=values_obj, path=new_path[:], **kwargs)
                )
        elif not values_obj:
            return matches
        else:
            for value_obj in values_obj:
                value = value_obj["Value"]
                child_path = value_obj["Path"]
                if not isinstance(value, dict):
                    if check_value:
                        matches.extend(
                            check_value(
                                value=value, path=new_path[:] + child_path, **kwargs
                            )
                        )
                else:
                    if len(value) == 1:
                        for dict_name, _ in value.items():
                            # If this is a function we shouldn't fall back to a check_value check
                            if dict_name in cfnlint.helpers.FUNCTIONS:
                                # convert the function name from camel case to underscore
                                # Example: Fn::FindInMap becomes check_find_in_map
                                function_name = f'check_{camel_to_snake(dict_name.replace("Fn::", ""))}'
                                if function_name == "check_ref":
                                    if check_ref:
                                        matches.extend(
                                            check_ref(
                                                value=value.get("Ref"),
                                                path=new_path[:] + child_path + ["Ref"],
                                                parameters=self.get_parameters(),
                                                resources=self.get_resources(),
                                                **kwargs,
                                            )
                                        )
                                else:
                                    if locals().get(function_name):
                                        matches.extend(
                                            locals()[function_name](
                                                value=value.get(dict_name),
                                                path=new_path[:]
                                                + child_path
                                                + [dict_name],
                                                **kwargs,
                                            )
                                        )
                            else:
                                if check_value:
                                    matches.extend(
                                        check_value(
                                            value=value,
                                            path=new_path[:] + child_path,
                                            **kwargs,
                                        )
                                    )
                    else:
                        if check_value:
                            matches.extend(
                                check_value(
                                    value=value, path=new_path[:] + child_path, **kwargs
                                )
                            )

        return matches

    def is_resource_available(self, path, resource):
        """
        Compares a path to resource to see if its available
        Returns scenarios that may result in the resource doesn't exist
        Input:
            Path: An array that is a Path to the object being checked
            Resource: The resource being compared to
        Output:
            If the resource is available the result is an empty array []
            If the resource is not available you will get a an array of Condition Names
                and when that condition is True or False will result in the resource
                not being available when trying to be associated.
                [{'ConditionName'}: False]
        """
        results = []
        path_conditions = self.get_conditions_from_path(self.template, path)
        resource_condition = (
            self.template.get("Resources", {}).get(resource, {}).get("Condition")
        )
        if resource_condition:
            # resource conditions are always true.  If the same resource condition exists in the path
            # with the same True value then nothing else matters
            test_path_conditions = copy(path_conditions)
            if not test_path_conditions.get(resource_condition):
                test_path_conditions[resource_condition] = set()
                if True not in test_path_conditions.get(resource_condition):
                    scenarios = self.conditions.get_scenarios(
                        test_path_conditions.keys()
                    )
                    for scenario in scenarios:
                        # We care about when the resource condition is false but the REF would still exist
                        if not scenario.get(resource_condition):
                            scenario_count = 0
                            for (
                                path_condition,
                                path_values,
                            ) in test_path_conditions.items():
                                if scenario.get(path_condition) in path_values:
                                    scenario_count += 1

                            if scenario_count == len(path_conditions):
                                results.append(scenario)

        # if resource condition isn't available then the resource is available
        return results

    def get_object_without_nested_conditions(self, obj, path):
        """
        Get a list of object values without conditions included.
        Evaluates deep into the object removing any nested conditions as well
        """
        results = []
        scenarios = self.get_condition_scenarios_below_path(path)
        if not isinstance(obj, (dict, list)):
            return results

        if not scenarios:
            if isinstance(obj, dict):
                if len(obj) == 1:
                    if obj.get("Ref") == "AWS::NoValue":
                        return results
            return [{"Scenario": None, "Object": obj}]

        def get_value(value, scenario):  # pylint: disable=R0911
            """Get the value based on the scenario resolving nesting"""
            if isinstance(value, dict):
                if len(value) == 1:
                    if "Fn::If" in value:
                        if_values = value.get("Fn::If")
                        if len(if_values) == 3:
                            if_path = scenario.get(if_values[0], None)
                            if if_path is not None:
                                if if_path:
                                    return get_value(if_values[1], scenario)
                                return get_value(if_values[2], scenario)
                    elif value.get("Ref") == "AWS::NoValue":
                        return None

                new_object = {}
                for k, v in value.items():
                    new_object[k] = get_value(v, scenario)
                return new_object
            if isinstance(value, list):
                new_list = []
                for item in value:
                    new_value = get_value(item, scenario)
                    if new_value is not None:
                        new_list.append(get_value(item, scenario))

                return new_list

            return value

        for scenario in scenarios:
            results.append({"Scenario": scenario, "Object": get_value(obj, scenario)})

        return results

    def get_value_from_scenario(self, obj, scenario):
        """
        Get object values from a provided scenario
        """

        def get_value(value, scenario):  # pylint: disable=R0911
            """Get the value based on the scenario resolving nesting"""
            if isinstance(value, dict):
                if len(value) == 1:
                    if "Fn::If" in value:
                        if_values = value.get("Fn::If")
                        if len(if_values) == 3:
                            if_path = scenario.get(if_values[0], None)
                            if if_path is not None:
                                if if_path:
                                    return get_value(if_values[1], scenario)
                                return get_value(if_values[2], scenario)
                    elif value.get("Ref") == "AWS::NoValue":
                        return None
                    else:
                        return value

                return value
            if isinstance(value, list):
                new_list = []
                for item in value:
                    new_value = get_value(item, scenario)
                    if new_value is not None:
                        new_list.append(get_value(item, scenario))

                return new_list

            return value

        result = cfnlint.dict_node({}, obj.start_mark, obj.end_mark)
        if isinstance(obj, dict):
            if len(obj) == 1:
                if obj.get("Fn::If"):
                    new_value = get_value(obj, scenario)
                    if new_value is not None:
                        result = new_value
                else:
                    for key, value in obj.items():
                        new_value = get_value(value, scenario)
                        if new_value is not None:
                            result[key] = new_value
            else:
                for key, value in obj.items():
                    new_value = get_value(value, scenario)
                    if new_value is not None:
                        result[key] = new_value

        return result

    def get_object_without_conditions(self, obj, property_names=None):
        """
        Gets a list of object values without conditions included
        Input:
            obj: The object/dict that makes up a set of properties
            Example:
                {
                    "DBSnapshotIdentifier" : {
                        "Fn::If" : [
                            "UseDBSnapshot",
                            {"Ref" : "DBSnapshotName"},
                            {"Ref" : "AWS::NoValue"}
                        ]
                    }
                }
        Output:
            A list of objects with scenarios for the conditions played out.
            If Ref to AWS::NoValue remove the property
            Example: [
                {
                    Object: {
                        "DBSnapshotIdentifier" : {"Ref" : "DBSnapshotName"}
                    },
                    Scenario: {UseDBSnapshot: True}
                }, {
                    Object: {
                    },
                    Scenario: {UseDBSnapshot: False}
                }
            ]
        """
        property_names = [] if property_names is None else property_names
        o = {}
        if property_names:
            for property_name in property_names:
                o[property_name] = deepcopy(obj.get(property_name))
        else:
            o = deepcopy(obj)
        results = []

        scenarios = self.get_conditions_scenarios_from_object([o])
        if not isinstance(obj, dict):
            return results

        if not scenarios:
            if isinstance(obj, dict):
                if len(obj) == 1:
                    if obj.get("Ref") == "AWS::NoValue":
                        return results
            return [{"Scenario": None, "Object": obj}]

        for scenario in scenarios:
            result_obj = self.get_value_from_scenario(obj, scenario)
            if result_obj:
                results.append({"Scenario": scenario, "Object": result_obj})

        return results

    def get_condition_scenarios_below_path(self, path, include_if_in_function=False):
        """
        get Condition Scenarios from below path
        """
        fn_ifs = self.search_deep_keys("Fn::If")
        results = {}
        for fn_if in fn_ifs:
            if len(fn_if) >= len(path):
                if path == fn_if[0 : len(path)]:
                    # This needs to handle items only below the Path
                    result = self.get_conditions_from_path(
                        self.template, fn_if[0:-1], False, include_if_in_function
                    )
                    for condition_name, condition_values in result.items():
                        if condition_name in results:
                            results[condition_name].union(condition_values)
                        else:
                            results[condition_name] = condition_values

        return self.conditions.get_scenarios(results.keys())

    def get_conditions_scenarios_from_object(self, objs):
        """
        Get condition from objects
        """

        def get_conditions_from_property(value):
            """Recursively get conditions"""
            results = set()
            if isinstance(value, dict):
                if len(value) == 1:
                    for k, v in value.items():
                        if k == "Fn::If":
                            if isinstance(v, list) and len(v) == 3:
                                if isinstance(v[0], str):
                                    results.add(v[0])
                                    results = results.union(
                                        get_conditions_from_property(v[1])
                                    )
                                    results = results.union(
                                        get_conditions_from_property(v[2])
                                    )
            elif isinstance(value, list):
                for v in value:
                    results = results.union(get_conditions_from_property(v))

            return results

        con = set()

        if isinstance(objs, dict):
            objs = [objs]

        for obj in objs:
            if isinstance(obj, dict):
                for k, v in obj.items():
                    # handle conditions directly under the object
                    if len(obj) == 1 and k == "Fn::If" and len(v) == 3:
                        con.add(v[0])
                        for r_c in v[1:]:
                            if isinstance(r_c, dict):
                                for s_k, s_v in r_c.items():
                                    if s_k == "Fn::If":
                                        con = con.union(
                                            get_conditions_from_property({s_k: s_v})
                                        )
                    else:
                        con = con.union(get_conditions_from_property(v))

        return self.conditions.get_scenarios(list(con))

    def get_conditions_from_path(
        self,
        text,
        path,
        include_resource_conditions=True,
        include_if_in_function=True,
        only_last=False,
    ):
        """
        Parent function to handle resources with conditions.
        Input:
            text: The object to start processing through the Path
            path: The path to recursively look for
        Output:
            An Object with keys being the Condition Names and the values are what
                if its in the True or False part of the path.
                {'condition': {True}}
        """

        results = self._get_conditions_from_path(
            text, path, include_if_in_function, only_last
        )
        if include_resource_conditions:
            if len(path) >= 2:
                if path[0] in ["Resources", "Outputs"]:
                    condition = text.get(path[0], {}).get(path[1], {}).get("Condition")
                    if condition:
                        if not results.get(condition):
                            results[condition] = set()
                        results[condition].add(True)

        return results

    def _get_conditions_from_path(
        self, text, path, include_if_in_function=True, only_last=False
    ):
        """
        Get the conditions and their True/False value for the path provided
        Input:
            text: The object to start processing through the Path
            path: The path to recursively look for
        Output:
            An Object with keys being the Condition Names and the values are what
                if its in the True or False part of the path.
                {'condition': {True}}
        """
        LOGGER.debug("Get conditions for path %s", path)
        results = {}

        def get_condition_name(value, num=None):
            """Test conditions for validity before providing the name"""
            con_path = set()
            if num == 1:
                con_path.add(True)
            elif num == 2:
                con_path.add(False)
            else:
                con_path = con_path.union((True, False))

            if value:
                if isinstance(value, list):
                    if len(value) == 3:
                        if not results.get(value[0]):
                            results[value[0]] = set()
                        results[value[0]] = results[value[0]].union(con_path)

        try:
            # Found a condition at the root of the Path
            if path[0] == "Fn::If" and (
                (len(path) == 1 and only_last) or not only_last
            ):
                condition = text.get("Fn::If")
                if len(path) > 1:
                    if path[1] in [1, 2]:
                        get_condition_name(condition, path[1])
                else:
                    get_condition_name(condition)
            # Iterate if the Path has more than one value
            if len(path) > 1:
                if (
                    path[0] in cfnlint.helpers.FUNCTIONS and path[0] != "Fn::If"
                ) and not include_if_in_function:
                    return results
                child_results = self._get_conditions_from_path(
                    text[path[0]], path[1:], include_if_in_function, only_last
                )
                for c_r_k, c_r_v in child_results.items():
                    if not results.get(c_r_k):
                        results[c_r_k] = set()
                    results[c_r_k] = results[c_r_k].union(c_r_v)

        except KeyError as _:
            pass

        return results


def camel_to_snake(s):
    """
    Is it ironic that this function is written in camel case, yet it
    converts to snake case? hmm..
    """
    _underscorer1 = re.compile(r"(.)([A-Z][a-z]+)")
    _underscorer2 = re.compile("([a-z0-9])([A-Z])")
    subbed = _underscorer1.sub(r"\1_\2", s)
    return _underscorer2.sub(r"\1_\2", subbed).lower()



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule
from cfnlint.rules.common import approaching_number_limit


class LimitNumber(CloudFormationLintRule):
    """Check maximum Parameter limit"""

    id = "I2010"
    shortdesc = "Parameter limit"
    description = (
        "Check the number of Parameters in the template is approaching the upper limit"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/cloudformation-limits.html"
    tags = ["parameters", "limits"]

    def match(self, cfn):
        return approaching_number_limit(cfn, "Parameters")



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Modules(CloudFormationLintRule):
    """Check that Modules do not contain invalid data"""

    id = "E5001"
    shortdesc = "Check that Modules resources are valid"
    description = "Check that Modules resources are valid"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/"
    tags = ["resources", "modules"]

    def match(self, cfn):
        matches = []
        resource_properties = cfn.get_resources()
        resource_dict = {
            key: resource_properties[key]
            for key in resource_properties
            if isinstance(resource_properties[key], dict)
        }
        for resource_name, resource_values in resource_dict.items():
            module = {
                "Type": v
                for (k, v) in resource_values.items()
                if str(v).endswith("::MODULE") is True
            }
            if module:
                matches.extend(self.check_metadata_keys(cfn))

                matches.extend(self.check_tags(resource_name, resource_values))

                matches.extend(
                    self.check_policy("CreationPolicy", resource_name, resource_values)
                )

                matches.extend(
                    self.check_policy("UpdatePolicy", resource_name, resource_values)
                )

        return matches

    def check_policy(self, policy, resource_name, resource_values):
        """Ensure invalid policies are not used"""
        matches = []
        if resource_values.get(policy, {}):
            path = ["Resources", resource_name, policy]
            matches.append(RuleMatch(path, f"{policy} is not permitted within Modules"))

        return matches

    def check_tags(self, resource_name, resource_values):
        """Ensure invalid policies are not used"""
        matches = []
        properties = resource_values.get("Properties", {})
        if properties.get("Tags"):
            path = ["Resources", resource_name, "Properties", "Tags"]
            matches.append(RuleMatch(path, "Tags is not permitted within Modules"))

        return matches

    def check_metadata_keys(self, cfn):
        """Ensure reserved metadata key AWS::CloudFormation::Module is not used"""
        modules = cfn.get_modules().keys()
        matches = []
        reserved_key = "AWS::CloudFormation::Module"
        refs = cfn.search_deep_keys(reserved_key)
        for ref in refs:
            if (ref[1] in modules) and (len(ref) > 3):
                if ref[0] == "Resources" and ref[2] == "Metadata":
                    matches.append(
                        RuleMatch(ref, f"The Metadata key {reserved_key} is reserved")
                    )
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class SubNotJoin(CloudFormationLintRule):
    """Check if Join is being used with no join characters"""

    id = "I1022"
    shortdesc = "Use Sub instead of Join"
    description = (
        "Prefer a sub instead of Join when using a join delimiter that is empty"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-sub.html"
    tags = ["functions", "sub", "join"]

    def _check_element(self, element):
        if isinstance(element, dict):
            if len(element) == 1:
                for key, value in element.items():
                    if key in ["Fn::Sub"]:
                        if not isinstance(value, str):
                            return False
                    elif key not in ["Ref", "Fn::GetAtt"]:
                        return False

        return True

    def _check_elements(self, elements):
        for element in elements:
            if not self._check_element(element):
                return False

        return True

    def match(self, cfn):
        matches = []

        join_objs = cfn.search_deep_keys("Fn::Join")

        for join_obj in join_objs:
            if isinstance(join_obj[-1], list):
                join_operator = join_obj[-1][0]
                join_elements = join_obj[-1][1]
                if isinstance(join_operator, str):
                    if join_operator == "":
                        if isinstance(join_elements, list):
                            if self._check_elements(join_elements):
                                matches.append(
                                    RuleMatch(
                                        join_obj[0:-1],
                                        "Prefer using Fn::Sub over Fn::Join with an empty delimiter",
                                    )
                                )
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
# pylint: disable=cyclic-import
import json

import cfnlint.rules.custom.Operators


# pylint: disable=too-many-return-statements
def make_rule(line, lineNumber):
    """Object Maker Function"""

    def get_value(value):
        raw_value = value.strip()
        try:
            return json.loads(raw_value)
        except ValueError:
            return raw_value

    def set_arguments(argument, error):
        raw_value = argument.split(error)
        value = process_sets(raw_value[0])
        error_message = None
        if len(raw_value) > 1:
            error_message = raw_value[1].strip().strip('"')
        return value, error_message

    def process_sets(raw_value):
        if (
            len(raw_value) > 1
            and raw_value[0] == "["
            and (raw_value[-2] == "]" or raw_value[-1] == "]")
        ):
            raw_value = raw_value[1:-2]
            raw_value = raw_value.split(",")
            for x in raw_value:
                x = x.strip()
            return raw_value

        return raw_value

    line = line.rstrip()
    rule_id = lineNumber + 9000
    line = line.split(" ", 3)
    error_level = "E"
    if len(line) == 4:
        resourceType = line[0]
        prop = line[1]
        operator = line[2]
        value = None
        error_message = None
        if "WARN" in line[3]:
            error_level = "W"
            value, error_message = set_arguments(line[3], "WARN")
        elif "ERROR" in line[3]:
            error_level = "E"
            value, error_message = set_arguments(line[3], "ERROR")
        else:
            value = process_sets(line[3])
            value = get_value(line[3])

        if isinstance(value, str):
            value = value.strip().strip('"')

        if operator in ["EQUALS", "=="]:
            return cfnlint.rules.custom.Operators.CreateEqualsRule(
                error_level + str(rule_id), resourceType, prop, value, error_message
            )
        if operator in ["NOT_EQUALS", "!="]:
            return cfnlint.rules.custom.Operators.CreateNotEqualsRule(
                error_level + str(rule_id), resourceType, prop, value, error_message
            )
        if operator == "IN":
            return cfnlint.rules.custom.Operators.CreateInSetRule(
                error_level + str(rule_id), resourceType, prop, value, error_message
            )
        if operator == "NOT_IN":
            return cfnlint.rules.custom.Operators.CreateNotInSetRule(
                error_level + str(rule_id), resourceType, prop, value, error_message
            )
        if operator == ">=":
            return cfnlint.rules.custom.Operators.CreateGreaterRule(
                error_level + str(rule_id), resourceType, prop, value, error_message
            )
        if operator == "<=":
            return cfnlint.rules.custom.Operators.CreateLesserRule(
                error_level + str(rule_id), resourceType, prop, value, error_message
            )

    return cfnlint.rules.custom.Operators.CreateInvalidRule(
        "E" + str(rule_id), operator
    )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import json

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class EqualsIsUseful(CloudFormationLintRule):
    """Validate that the Equals will return true/false and not always be true or false"""

    id = "W8003"
    shortdesc = "Fn::Equals will always return true or false"
    description = "Validate Fn::Equals to see if its comparing two strings or two equal items. While this works it may not be intended."
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-conditions.html#intrinsic-function-reference-conditions-equals"
    tags = ["functions", "equals"]

    function = "Fn::Equals"

    def _check_equal_values(self, values, path):
        matches = []

        if json.dumps(values[0]) == json.dumps(values[1]):
            message = self.function + " element will alway return true"
            matches.append(RuleMatch(path, message))
        elif isinstance(values[0], str) and isinstance(values[1], str):
            message = self.function + " element will alway return false"
            matches.append(RuleMatch(path, message))
        return matches

    def match(self, cfn):
        matches = []
        # Build the list of functions
        trees = cfn.search_deep_keys(self.function)

        for tree in trees:
            # Test when in Conditions
            if tree[0] == "Conditions":
                value = tree[-1]
                if isinstance(value, list) and len(value) == 2:
                    matches.extend(self._check_equal_values(value, tree[:-1]))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import json

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class StateMachine(CloudFormationLintRule):
    """Check State Machine Definition"""

    id = "E2532"
    shortdesc = "Check State Machine Definition for proper syntax"
    description = (
        "Check the State Machine String Definition to make sure its JSON. "
        "Validate basic syntax of the file to determine validity."
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-stepfunctions-statemachine.html"
    tags = ["resources", "stepfunctions"]

    def __init__(self):
        """Init"""
        super().__init__()
        self.resource_property_types.append("AWS::StepFunctions::StateMachine")

    def _check_state_json(self, def_json, state_name, path):
        """Check State JSON Definition"""
        matches = []

        # https://docs.aws.amazon.com/step-functions/latest/dg/amazon-states-language-common-fields.html
        common_state_keys = [
            "Next",
            "End",
            "Type",
            "Comment",
            "InputPath",
            "OutputPath",
        ]
        common_state_required_keys = [
            "Type",
        ]
        state_key_types = {
            "Pass": ["Result", "ResultPath", "Parameters"],
            "Task": [
                "Credentials",
                "Resource",
                "Parameters",
                "ResultPath",
                "ResultSelector",
                "Retry",
                "Catch",
                "TimeoutSeconds",
                "TimeoutSecondsPath",
                "Parameters",
                "HeartbeatSeconds",
                "HeartbeatSecondsPath",
            ],
            "Map": [
                "MaxConcurrency",
                "Iterator",
                "ItemsPath",
                "ItemProcessor",
                "ItemSelector",
                "ResultPath",
                "ResultSelector",
                "Retry",
                "Catch",
                "Parameters",
            ],
            "Choice": ["Choices", "Default"],
            "Wait": ["Seconds", "Timestamp", "SecondsPath", "TimestampPath"],
            "Succeed": [],
            "Fail": ["Cause", "Error"],
            "Parallel": [
                "Branches",
                "ResultPath",
                "ResultSelector",
                "Parameters",
                "Retry",
                "Catch",
            ],
        }
        state_required_types = {
            "Pass": [],
            "Task": ["Resource"],
            "Choice": ["Choices"],
            "Wait": [],
            "Succeed": [],
            "Fail": [],
            "Parallel": ["Branches"],
        }

        for req_key in common_state_required_keys:
            if req_key not in def_json:
                message = f"State Machine Definition required key ({req_key}) for State ({state_name}) is missing"
                matches.append(RuleMatch(path, message))
                return matches

        state_type = def_json.get("Type")

        if state_type in state_key_types:
            for state_key, _ in def_json.items():
                if state_key not in common_state_keys + state_key_types.get(
                    state_type, []
                ):
                    message = f"State Machine Definition key ({state_key}) for State ({state_name}) of Type ({state_type}) is not valid"
                    matches.append(RuleMatch(path, message))
            for req_key in common_state_required_keys + state_required_types.get(
                state_type, []
            ):
                if req_key not in def_json:
                    message = f"State Machine Definition required key ({req_key}) for State ({state_name}) of Type ({state_type}) is missing"
                    matches.append(RuleMatch(path, message))
                    return matches
        else:
            message = f"State Machine Definition Type ({state_type}) is not valid"
            matches.append(RuleMatch(path, message))

        return matches

    def _check_definition_json(self, def_json, path):
        """Check JSON Definition"""
        matches = []

        top_level_keys = ["Comment", "StartAt", "TimeoutSeconds", "Version", "States"]
        top_level_required_keys = ["StartAt", "States"]
        for top_key, _ in def_json.items():
            if top_key not in top_level_keys:
                message = f"State Machine Definition key ({top_key}) is not valid"
                matches.append(RuleMatch(path, message))

        for req_key in top_level_required_keys:
            if req_key not in def_json:
                message = (
                    f"State Machine Definition required key ({req_key}) is missing"
                )
                matches.append(RuleMatch(path, message))

        for state_name, state_value in def_json.get("States", {}).items():
            matches.extend(self._check_state_json(state_value, state_name, path))
        return matches

    def check_value(self, value, path, fail_on_loads=True):
        """Check Definition Value"""
        matches = []
        try:
            def_json = json.loads(value)
        # pylint: disable=W0703
        except Exception as err:
            if fail_on_loads:
                message = f"State Machine Definition needs to be formatted as JSON. Error {err}"
                matches.append(RuleMatch(path, message))
                return matches

            self.logger.debug("State Machine definition could not be parsed. Skipping")
            return matches

        matches.extend(self._check_definition_json(def_json, path))
        return matches

    def check_sub(self, value, path):
        """Check Sub Object"""
        matches = []
        if isinstance(value, list):
            matches.extend(self.check_value(value[0], path, False))
        elif isinstance(value, str):
            matches.extend(self.check_value(value, path, False))

        return matches

    def match_resource_properties(self, properties, _, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        matches.extend(
            cfn.check_value(
                obj=properties,
                key="DefinitionString",
                path=path[:],
                check_value=self.check_value,
                check_sub=self.check_sub,
            )
        )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.helpers import VALID_PARAMETER_TYPES
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Types(CloudFormationLintRule):
    """Check if Parameters are typed"""

    id = "E2002"
    shortdesc = "Parameters have appropriate type"
    description = "Making sure the parameters have a correct type"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/best-practices.html#parmtypes"
    tags = ["parameters"]

    def match(self, cfn):
        matches = []

        for paramname, paramvalue in cfn.get_parameters_valid().items():
            # If the type isn't found we create a valid one
            # this test isn't about missing required properties for a
            # parameter.
            paramtype = paramvalue.get("Type", "String")
            if paramtype not in VALID_PARAMETER_TYPES:
                message = "Parameter {0} has invalid type {1}"
                matches.append(
                    RuleMatch(
                        ["Parameters", paramname, "Type"],
                        message.format(paramname, paramtype),
                    )
                )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from datetime import datetime

from cfnlint.data import AdditionalSpecs
from cfnlint.helpers import load_resource
from cfnlint.rules import CloudFormationLintRule


class DeprecatedRuntime(CloudFormationLintRule):
    """Check if EOL Lambda Function Runtimes are used"""

    def __init__(self):
        """Init"""
        super().__init__()
        self.resource_property_types.append("AWS::Lambda::Function")
        self.deprecated_runtimes = load_resource(
            AdditionalSpecs, "LmbdRuntimeLifecycle.json"
        )

    current_date = datetime.today()

    def check_runtime(self, runtime_value, path):
        """Check if the given runtime is valid"""
        self.logger.debug(runtime_value, path)
        return []

    def check_value(self, value, path):
        """Check Lambda Runtime value"""
        matches = []
        matches.extend(self.check_runtime(value, path))
        return matches

    def check_ref(self, value, path, parameters, resources):  # pylint: disable=W0613
        """Check Lambda Runtime Ref value"""
        matches = []
        if value in parameters:
            parameter = parameters.get(value, {})
            param_path = ["Parameters", value]

            # Validate the Default
            default_value = parameter.get("Default", "")
            matches.extend(self.check_runtime(default_value, param_path + ["Default"]))

            # Validate AllowedValues
            allowed_values = parameter.get("AllowedValues")
            if isinstance(allowed_values, list):
                param_path = param_path + ["AllowedValues"]
                for index, allowed_value in enumerate(allowed_values):
                    matches.extend(
                        self.check_runtime(allowed_value, param_path + [index])
                    )

        return matches

    def match_resource_properties(self, properties, _, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        matches.extend(
            cfn.check_value(
                obj=properties,
                key="Runtime",
                path=path[:],
                check_value=self.check_value,
                check_ref=self.check_ref,
            )
        )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.helpers import RESOURCE_SPECS, VALID_PARAMETER_TYPES_LIST
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Join(CloudFormationLintRule):
    """Check if Join values are correct"""

    id = "E1022"
    shortdesc = "Join validation of parameters"
    description = "Making sure the join function is properly configured"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-join.html"
    tags = ["functions", "join"]

    def __init__(self):
        """Initialize the rule"""
        super().__init__()
        self.list_supported_functions = []
        self.singular_supported_functions = []
        for intrinsic_type, intrinsic_value in (
            RESOURCE_SPECS.get("us-east-1").get("IntrinsicTypes").items()
        ):
            if "List" in intrinsic_value.get("ReturnTypes", []):
                self.list_supported_functions.append(intrinsic_type)
            if "Singular" in intrinsic_value.get("ReturnTypes", []):
                self.singular_supported_functions.append(intrinsic_type)

    def _get_parameters(self, cfn):
        """Get all Parameter Names"""
        results = {}
        parameters = cfn.template.get("Parameters", {})
        if isinstance(parameters, dict):
            for param_name, param_values in parameters.items():
                # This rule isn't here to check the Types but we need
                # something valid if it doesn't exist
                results[param_name] = param_values.get("Type", "String")

        return results

    def _normalize_getatt(self, getatt):
        """Normalize getatt into an array"""
        if isinstance(getatt, str):
            return getatt.split(".", 1)
        return getatt

    def _is_ref_a_list(self, parameter, template_parameters):
        """Is a Ref a list"""
        list_params = [
            "AWS::NotificationARNs",
        ]

        if parameter in template_parameters:
            if template_parameters.get(parameter) in VALID_PARAMETER_TYPES_LIST:
                return True
        if parameter in list_params:
            return True
        return False

    def _is_getatt_a_list(self, parameter, get_atts):
        """Is a GetAtt a List"""

        for resource, attributes in get_atts.items():
            for attribute_name, attribute_values in attributes.items():
                if resource == parameter[0] and attribute_name == "*":
                    if attribute_values.get("PrimitiveItemType"):
                        return "FALSE"
                    if attribute_values.get("Type") == "List":
                        return "TRUE"
                    return "UNKNOWN"
                if resource == parameter[0] and attribute_name == parameter[1]:
                    if attribute_values.get("Type") == "List":
                        return "TRUE"

        return "FALSE"

    def _match_string_objs(self, join_string_objs, cfn, path):
        """Check join list"""

        matches = []

        template_parameters = self._get_parameters(cfn)
        get_atts = cfn.get_valid_getatts()

        if isinstance(join_string_objs, dict):
            if len(join_string_objs) == 1:
                for key, value in join_string_objs.items():
                    if key not in self.list_supported_functions:
                        message = "Fn::Join unsupported function for {0}"
                        matches.append(
                            RuleMatch(path, message.format("/".join(map(str, path))))
                        )
                    elif key in ["Ref"]:
                        if not self._is_ref_a_list(value, template_parameters):
                            message = "Fn::Join must use a list at {0}"
                            matches.append(
                                RuleMatch(
                                    path, message.format("/".join(map(str, path)))
                                )
                            )
                    elif key in ["Fn::GetAtt"]:
                        if (
                            self._is_getatt_a_list(
                                self._normalize_getatt(value), get_atts
                            )
                            == "FALSE"
                        ):
                            message = "Fn::Join must use a list at {0}"
                            matches.append(
                                RuleMatch(
                                    path, message.format("/".join(map(str, path)))
                                )
                            )
            else:
                message = "Join list of values should be singular for {0}"
                matches.append(
                    RuleMatch(path, message.format("/".join(map(str, path))))
                )
        elif not isinstance(join_string_objs, list):
            message = "Join list of values for {0}"
            matches.append(RuleMatch(path, message.format("/".join(map(str, path)))))
        else:
            for string_obj in join_string_objs:
                if isinstance(string_obj, dict):
                    if len(string_obj) == 1:
                        for key, value in string_obj.items():
                            if key not in self.singular_supported_functions:
                                message = "Join unsupported function for {0}"
                                matches.append(
                                    RuleMatch(
                                        path, message.format("/".join(map(str, path)))
                                    )
                                )
                            elif key in ["Ref"]:
                                if self._is_ref_a_list(value, template_parameters):
                                    message = "Fn::Join must not be a list at {0}"
                                    matches.append(
                                        RuleMatch(
                                            path,
                                            message.format("/".join(map(str, path))),
                                        )
                                    )
                            elif key in ["Fn::GetAtt"]:
                                if (
                                    self._is_getatt_a_list(
                                        self._normalize_getatt(value), get_atts
                                    )
                                    == "TRUE"
                                ):
                                    message = "Fn::Join must not be a list at {0}"
                                    matches.append(
                                        RuleMatch(
                                            path,
                                            message.format("/".join(map(str, path))),
                                        )
                                    )
                    else:
                        message = "Join list of values should be singular for {0}"
                        matches.append(
                            RuleMatch(path, message.format("/".join(map(str, path))))
                        )
                elif not isinstance(string_obj, str):
                    message = "Join list of singular function or string for {0}"
                    matches.append(
                        RuleMatch(path, message.format("/".join(map(str, path))))
                    )

        return matches

    def match(self, cfn):
        matches = []

        join_objs = cfn.search_deep_keys("Fn::Join")

        for join_obj in join_objs:
            join_value_obj = join_obj[-1]
            path = join_obj[:-1]
            if isinstance(join_value_obj, list):
                if len(join_value_obj) == 2:
                    join_string = join_value_obj[0]
                    join_string_objs = join_value_obj[1]
                    if not isinstance(join_string, str):
                        message = "Join string has to be of type string for {0}"
                        matches.append(
                            RuleMatch(path, message.format("/".join(map(str, path))))
                        )
                    matches.extend(self._match_string_objs(join_string_objs, cfn, path))
                else:
                    message = "Join should be an array of 2 for {0}"
                    matches.append(
                        RuleMatch(path, message.format("/".join(map(str, path))))
                    )
            else:
                message = "Join should be an array of 2 for {0}"
                matches.append(
                    RuleMatch(path, message.format("/".join(map(str, path))))
                )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from typing import List

from cfnlint.config import configure_logging
from cfnlint.core import get_rules
from cfnlint.decode import decode_str
from cfnlint.helpers import REGIONS
from cfnlint.rules import Match, RulesCollection
from cfnlint.runner import Runner

Matches = List[Match]


def lint(s: str, rules: RulesCollection, regions: List[str]) -> Matches:
    """Validate a string template using the specified rules and regions.

    Parameters
    ----------
    s : str
        the template string
    rules : RulesCollection
        The rules to run against s
    regions : List[str]
        The regions to test against s

    Returns
    -------
    list
        a list of errors if any were found, else an empty list
    """
    configure_logging(None, None)
    template, errors = decode_str(s)
    if errors:
        return errors

    if template is None:
        return []

    runner = Runner(
        rules=rules,
        filename=None,
        template=template,
        regions=regions,
        verbosity=0,
        mandatory_rules=None,
    )
    return runner.run()


def lint_all(s: str) -> Matches:
    """Validate a string template against all regions and rules.

    Parameters
    ----------
    s : str
        the template string

    Returns
    -------
    list
        a list of errors if any were found, else an empty list
    """
    return lint(
        s=s,
        rules=get_rules([], [], ["I", "W", "E"], include_experimental=True),
        regions=REGIONS,
    )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import logging
import sys
import warnings

import cfnlint.core

LOGGER = logging.getLogger("cfnlint")


def main():
    if sys.version_info[:2] == (3, 4):
        warnings.warn(
            "Python 3.4 has reached end of life. "
            "cfn-lint has ended support for python 3.4 as of July 1st, 2020.",
            Warning,
            stacklevel=3,
        )
    elif sys.version_info[:2] == (3, 5):
        warnings.warn(
            "Python 3.5 has reached end of life. "
            "cfn-lint has ended support for python 3.5 as of August 1st, 2021.",
            Warning,
            stacklevel=3,
        )
    elif sys.version_info[:2] == (2, 7):
        warnings.warn(
            "Python 2.7 has reached end of life. "
            "cfn-lint will end support for python 2.7 on June 1st, 2020.",
            Warning,
            stacklevel=3,
        )

    try:
        (args, filenames, formatter) = cfnlint.core.get_args_filenames(sys.argv[1:])
        matches = list(cfnlint.core.get_matches(filenames, args))
        rules = cfnlint.core.get_used_rules()
        matches_output = formatter.print_matches(matches, rules, filenames)

        if matches_output:
            if args.output_file:
                with open(args.output_file, "w", encoding="utf-8") as output_file:
                    output_file.write(matches_output)
            else:
                print(matches_output)

        return cfnlint.core.get_exit_code(matches, args.non_zero_exit_code)
    except cfnlint.core.CfnLintExitException as e:
        LOGGER.error(str(e))
        return e.exit_code


if __name__ == "__main__":
    try:
        sys.exit(main())
    except (ValueError, TypeError):
        LOGGER.error(ValueError)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class DomainValidationOptions(CloudFormationLintRule):
    """Check if a certificate's domain validation options are set up correctly"""

    id = "E3503"
    shortdesc = "ValidationDomain is superdomain of DomainName"
    description = "In ValidationDomainOptions, the ValidationDomain must be a superdomain of the DomainName being validated"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-certificatemanager-certificate-domainvalidationoption.html#cfn-certificatemanager-certificate-domainvalidationoption-validationdomain"
    tags = [
        "certificate",
        "certificatemanager",
        "domainvalidationoptions",
        "validationdomain",
    ]

    def __init__(self):
        """Init"""
        super().__init__()
        self.resource_property_types = ["AWS::CertificateManager::Certificate"]

    def check_value(self, value, path, **kwargs):
        """Check value inside the list of DomainValidationOptions"""
        matches = []
        cfn = kwargs.get("cfn")
        if isinstance(value, dict):
            property_sets = cfn.get_object_without_conditions(value)
            for property_set in property_sets:
                properties = property_set.get("Object")
                scenario = property_set.get("Scenario")
                domain_name = properties.get("DomainName", None)
                validation_domain = properties.get("ValidationDomain", None)
                if isinstance(domain_name, str) and isinstance(validation_domain, str):
                    if domain_name == validation_domain:
                        continue

                    if not domain_name.endswith("." + validation_domain):
                        message = (
                            "ValidationDomain must be a superdomain of DomainName at {}"
                        )
                        if scenario is None:
                            matches.append(
                                RuleMatch(
                                    path[:] + ["DomainName"],
                                    message.format("/".join(map(str, path))),
                                )
                            )
                        else:
                            scenario_text = " and ".join(
                                [
                                    f'when condition "{k}" is {v}'
                                    for (k, v) in scenario.items()
                                ]
                            )
                            matches.append(
                                RuleMatch(
                                    path[:] + ["DomainName"],
                                    message.format(
                                        "/".join(map(str, path)) + " " + scenario_text
                                    ),
                                )
                            )
        return matches

    def match_resource_properties(self, properties, _, path, cfn):
        matches = []
        matches.extend(
            cfn.check_value(
                properties,
                "DomainValidationOptions",
                path[:],
                check_value=self.check_value,
                cfn=cfn,
            )
        )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Select(CloudFormationLintRule):
    """Check if Select values are correct"""

    id = "E1017"
    shortdesc = "Select validation of parameters"
    description = "Making sure the Select function is properly configured"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-select.html"
    tags = ["functions", "select"]

    def match(self, cfn):
        matches = []

        select_objs = cfn.search_deep_keys("Fn::Select")

        supported_functions = [
            "Fn::FindInMap",
            "Fn::GetAtt",
            "Fn::GetAZs",
            "Fn::If",
            "Fn::Split",
            "Fn::Cidr",
            "Ref",
        ]

        for select_obj in select_objs:
            select_value_obj = select_obj[-1]
            tree = select_obj[:-1]
            if isinstance(select_value_obj, list):
                if len(select_value_obj) == 2:
                    index_obj = select_value_obj[0]
                    list_of_objs = select_value_obj[1]
                    if isinstance(index_obj, dict):
                        if len(index_obj) == 1:
                            for index_key, _ in index_obj.items():
                                if index_key not in [
                                    "Ref",
                                    "Fn::FindInMap",
                                    "Fn::Select",
                                ]:
                                    message = "Select index should be an Integer or a function Ref or FindInMap for {0}"
                                    matches.append(
                                        RuleMatch(
                                            tree,
                                            message.format("/".join(map(str, tree))),
                                        )
                                    )
                    elif not isinstance(index_obj, int):
                        try:
                            int(index_obj)
                        except ValueError:
                            message = "Select index should be an Integer or a function of Ref or FindInMap for {0}"
                            matches.append(
                                RuleMatch(
                                    tree, message.format("/".join(map(str, tree)))
                                )
                            )
                    if isinstance(list_of_objs, dict):
                        if len(list_of_objs) == 1:
                            for key, _ in list_of_objs.items():
                                if key not in supported_functions:
                                    message = (
                                        "Select should use a supported function of {0}"
                                    )
                                    matches.append(
                                        RuleMatch(
                                            tree,
                                            message.format(
                                                ", ".join(map(str, supported_functions))
                                            ),
                                        )
                                    )
                        else:
                            message = "Select should use a supported function of {0}"
                            matches.append(
                                RuleMatch(
                                    tree,
                                    message.format(
                                        ", ".join(map(str, supported_functions))
                                    ),
                                )
                            )
                    elif not isinstance(list_of_objs, list):
                        message = "Select should be an array of values for {0}"
                        matches.append(
                            RuleMatch(tree, message.format("/".join(map(str, tree))))
                        )
                else:
                    message = "Select should be a list of 2 elements for {0}"
                    matches.append(
                        RuleMatch(tree, message.format("/".join(map(str, tree))))
                    )
            else:
                message = "Select should be a list of 2 elements for {0}"
                matches.append(
                    RuleMatch(tree, message.format("/".join(map(str, tree))))
                )
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class FindInMapKeys(CloudFormationLintRule):
    """Check if FindInMap values are correct"""

    id = "W1011"
    shortdesc = "FindInMap keys exist in the map"
    description = (
        "Checks the keys in a FindInMap to make sure they exist. "
        "Check only if the Map Name is a string and if the key is a string."
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-findinmap.html"
    tags = ["functions", "findinmap"]

    def check_keys(self, map_name, keys, mappings, tree):
        """Check the validity of the first key"""
        matches = []
        first_key = keys[0]
        second_key = keys[1]
        if isinstance(second_key, (str, int)):
            if isinstance(map_name, (str)):
                mapping = mappings.get(map_name)
                if mapping:
                    if isinstance(first_key, (str, int)):
                        if isinstance(map_name, (str)):
                            if mapping.get(first_key) is None:
                                message = 'FindInMap first key "{0}" doesn\'t exist in map "{1}" at {3}'
                                matches.append(
                                    RuleMatch(
                                        tree[:] + [1],
                                        message.format(
                                            first_key,
                                            map_name,
                                            first_key,
                                            "/".join(map(str, tree)),
                                        ),
                                    )
                                )
                        if mapping.get(first_key):
                            # Don't double error if they first key doesn't exist
                            if mapping.get(first_key, {}).get(second_key) is None:
                                message = 'FindInMap second key "{0}" doesn\'t exist in map "{1}" under "{2}" at {3}'
                                matches.append(
                                    RuleMatch(
                                        tree[:] + [2],
                                        message.format(
                                            second_key,
                                            map_name,
                                            first_key,
                                            "/".join(map(str, tree)),
                                        ),
                                    )
                                )
                    else:
                        for key, value in mapping.items():
                            if value.get(second_key) is None:
                                message = 'FindInMap second key "{0}" doesn\'t exist in map "{1}" under "{2}" at {3}'
                                matches.append(
                                    RuleMatch(
                                        tree[:] + [2],
                                        message.format(
                                            second_key,
                                            map_name,
                                            key,
                                            "/".join(map(str, tree)),
                                        ),
                                    )
                                )

        return matches

    def match(self, cfn):
        matches = []

        findinmaps = cfn.search_deep_keys("Fn::FindInMap")
        mappings = cfn.get_mappings()
        for findinmap in findinmaps:
            tree = findinmap[:-1]
            map_obj = findinmap[-1]

            if len(map_obj) == 3:
                matches.extend(self.check_keys(map_obj[0], map_obj[1:], mappings, tree))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.helpers import RESOURCE_SPECS
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class NumberSize(CloudFormationLintRule):
    """Check if a Number has a length within the limit"""

    id = "E3034"
    shortdesc = "Check if a number is between min and max"
    description = "Check numbers (integers and floats) for its value being between the minimum and maximum"
    source_url = "https://github.com/awslabs/cfn-python-lint/blob/main/docs/cfn-resource-specification.md#allowedpattern"
    tags = ["resources", "property", "number", "size"]

    def initialize(self, cfn):
        """Initialize the rule"""
        for resource_type_spec in RESOURCE_SPECS.get(cfn.regions[0]).get(
            "ResourceTypes"
        ):
            self.resource_property_types.append(resource_type_spec)
        for property_type_spec in RESOURCE_SPECS.get(cfn.regions[0]).get(
            "PropertyTypes"
        ):
            self.resource_sub_property_types.append(property_type_spec)

    def _check_number_value(self, value, path, **kwargs):
        """Check if the value is in the given ranges"""
        matches = []
        number_min = kwargs.get("number_min")
        number_max = kwargs.get("number_max")

        number_types = (
            float,
            int,
        )

        if isinstance(value, str):
            try:
                value = float(value)
            except ValueError:
                message = "Value has to be between {0} and {1} at {2}"
                matches.append(
                    RuleMatch(
                        path,
                        message.format(
                            number_min, number_max, "/".join(map(str, path))
                        ),
                    )
                )

        if isinstance(value, number_types):
            if not number_min <= value <= number_max:
                message = "Value has to be between {0} and {1} at {2}"
                matches.append(
                    RuleMatch(
                        path,
                        message.format(
                            number_min, number_max, "/".join(map(str, path))
                        ),
                    )
                )

        return matches

    def check(self, cfn, properties, specs, path):
        """Check itself"""
        matches = []
        for p_value, p_path in properties.items_safe(path[:]):
            for prop in p_value:
                if prop in specs:
                    value_type = specs.get(prop).get("Value", {}).get("ValueType", "")
                    if value_type:
                        property_type = specs.get(prop).get("PrimitiveType")
                        value_specs = (
                            RESOURCE_SPECS.get(cfn.regions[0])
                            .get("ValueTypes")
                            .get(value_type, {})
                        )
                        if value_specs == "CACHED":
                            value_specs = (
                                RESOURCE_SPECS.get("us-east-1")
                                .get("ValueTypes")
                                .get(value_type, {})
                            )
                        if (
                            value_specs.get("NumberMax") is not None
                            and value_specs.get("NumberMin") is not None
                        ):
                            if property_type in ["Integer", "Double", "Long", "String"]:
                                matches.extend(
                                    cfn.check_value(
                                        properties,
                                        prop,
                                        p_path,
                                        check_value=self._check_number_value,
                                        number_max=value_specs.get("NumberMax"),
                                        number_min=value_specs.get("NumberMin"),
                                    )
                                )
        return matches

    def match_resource_sub_properties(self, properties, property_type, path, cfn):
        """Match for sub properties"""
        matches = []

        specs = (
            RESOURCE_SPECS.get(cfn.regions[0])
            .get("PropertyTypes")
            .get(property_type, {})
            .get("Properties", {})
        )
        matches.extend(self.check(cfn, properties, specs, path))

        return matches

    def match_resource_properties(self, properties, resource_type, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        specs = (
            RESOURCE_SPECS.get(cfn.regions[0])
            .get("ResourceTypes")
            .get(resource_type, {})
            .get("Properties", {})
        )
        matches.extend(self.check(cfn, properties, specs, path))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class InterfaceConfiguration(CloudFormationLintRule):
    """Check if Metadata Interface Configuration are configured correctly"""

    id = "E4001"
    shortdesc = "Metadata Interface have appropriate properties"
    description = "Metadata Interface properties are properly configured"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-cloudformation-interface.html"
    tags = ["metadata"]

    valid_keys = ["ParameterGroups", "ParameterLabels"]

    def match(self, cfn):
        """Check CloudFormation Metadata Interface Configuration"""

        matches = []

        strinterface = "AWS::CloudFormation::Interface"

        metadata_obj = cfn.template.get("Metadata", {})
        if metadata_obj:
            interfaces = metadata_obj.get(strinterface, {})
            if isinstance(interfaces, dict):
                for interface in interfaces:
                    if interface not in self.valid_keys:
                        message = "Metadata Interface has invalid property {0}"
                        matches.append(
                            RuleMatch(
                                ["Metadata", strinterface, interface],
                                message.format(interface),
                            )
                        )
                parameter_groups = interfaces.get("ParameterGroups", [])
                for index, value in enumerate(parameter_groups):
                    for key in value:
                        if key not in ["Label", "Parameters"]:
                            message = "Metadata Interface has invalid property {0}"
                            matches.append(
                                RuleMatch(
                                    [
                                        "Metadata",
                                        strinterface,
                                        "ParameterGroups",
                                        index,
                                        key,
                                    ],
                                    message.format(key),
                                )
                            )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules.resources.properties.BasedOnValue import BasedOnValue


class RequiredBasedOnValue(BasedOnValue):
    """Check Required Properties are supplied when another property has a certain value"""

    id = "E3017"
    shortdesc = "Property is required based on another properties value"
    description = (
        "When certain properties have a certain value it results in other properties being required. "
        "This rule will validate those required properties are specified when those values are supplied"
    )
    tags = ["resources"]
    spec_type = "RequiredProperties"
    message = "should be specified"

    def _check_prop(self, prop, scenario):
        return prop not in scenario.get("Object")



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import codecs
import re
from setuptools import find_packages
from setuptools import setup


def get_version(filename):
    with codecs.open(filename, 'r', 'utf-8') as fp:
        contents = fp.read()
    return re.search(r"__version__ = ['\"]([^'\"]+)['\"]", contents).group(1)


version = get_version('src/cfnlint/version.py')


with open('README.md', encoding='utf-8') as f:
    readme = f.read()

setup(
    name='cfn-lint',
    version=version,
    description=('Checks CloudFormation templates for practices and behaviour \
that could potentially be improved'),
    long_description=readme,
    long_description_content_type="text/markdown",
    keywords='aws, lint',
    author='kddejong',
    author_email='kddejong@amazon.com',
    url='https://github.com/aws-cloudformation/cfn-python-lint',
    package_dir={'': 'src'},
    package_data={'cfnlint': [
        'data/CloudSpecs/*.json',
        'data/AdditionalSpecs/*.json',
        'data/Serverless/*.json',
        'data/ExtendedSpecs/*/*.json',
        'data/CfnLintCli/config/schema.json'
    ]},
    packages=find_packages('src'),
    zip_safe=False,
    install_requires=[
        'pyyaml>5.4',
        'aws-sam-translator>=1.60.1',
        'jsonpatch',
        'jsonschema>=3.0,<5',
        'networkx>=2.4,<4',
        'junit-xml~=1.9',
        'jschema_to_python~=1.2.3',
        'sarif-om~=1.0.4'
    ],
    python_requires='>=3.7, <=4.0, !=4.0',
    entry_points={
        'console_scripts': [
            'cfn-lint = cfnlint.__main__:main'
        ]
    },
    license='MIT no attribution',
    test_suite="unittest",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import hashlib
import json
import logging
from copy import copy

import cfnlint.helpers

LOGGER = logging.getLogger(__name__)


class ConditionParseError(Exception):
    """
    Error thrown when the template has poorly formatted condition
    """


def get_hash(obj):
    """Return a hash of an object"""
    return hashlib.sha1(json.dumps(obj, sort_keys=True).encode("utf-8")).hexdigest()


class EqualsValue:
    """holds the values of a equals"""

    Function = None
    String = None

    def __init__(self, value):
        if isinstance(value, dict):
            if len(value) == 1:
                # Save hashes of the dict for consistency and sorting
                self.Function = get_hash(value)
            else:
                LOGGER.debug("Length of the object needs to be 1")
                raise ConditionParseError
        elif isinstance(value, str):
            self.String = value
        elif isinstance(value, int):
            self.String = str(value)
        else:
            LOGGER.debug("Equals value has to be string or object")
            raise ConditionParseError

    def __eq__(self, other):
        return other in [self.Function, self.String]


class Equals:
    """Equals operator"""

    Left = None
    Right = None

    def __init__(self, equals):
        if isinstance(equals, list):
            if len(equals) == 2:
                self.Left = EqualsValue(equals[0])
                self.Right = EqualsValue(equals[1])
            else:
                LOGGER.debug("Length of Equals needs to be 2")
                raise ConditionParseError
        else:
            LOGGER.debug("Equals needs to be a list")
            raise ConditionParseError

    def test(self, scenarios):
        """Do an equals based on the provided scenario"""
        for scenario, value in scenarios.items():
            if scenario == self.Left:
                return value == self.Right
            if scenario == self.Right:
                return value == self.Left

        return None


class Condition:
    """Individual condition"""

    And = None
    Or = None
    Not = None
    Equals = None
    Influenced_Equals = None
    Name = None

    def __init__(self, template, name=None, sub_condition=None):
        self.And = []
        self.Or = []
        self.Not = []
        self.Influenced_Equals = {}
        self.Name = name
        if name is not None:
            value = template.get("Conditions", {}).get(name, {})
            try:
                self.process_condition(template, value)
            except ConditionParseError:
                LOGGER.debug("Error parsing condition: %s", name)
                self.Equals = None
        elif sub_condition is not None:
            try:
                self.process_condition(template, sub_condition)
            except ConditionParseError:
                LOGGER.debug("Error parsing condition: %s", name)
                self.Equals = None

    def __eq__(self, __o: object) -> bool:
        if self.Name is None:
            return False
        return self.Name == __o.Name

    def test(self, scenarios):
        """Test a condition based on a scenario"""
        if self.And:
            for a in self.And:
                if not a.test(scenarios):
                    return False
            return True
        if self.Or:
            for o in self.Or:
                if o.test(scenarios):
                    return True
            return False
        if self.Not:
            for n in self.Not:
                return not n.test(scenarios)
        return self.Equals.test(scenarios)

    def process_influenced_equal(self, equal):
        """Get influenced equals from sub conditions"""
        if equal.Left.Function:
            if not self.Influenced_Equals.get(equal.Left.Function):
                self.Influenced_Equals[equal.Left.Function] = set()
            if equal.Right.Function:
                self.Influenced_Equals[equal.Left.Function].add(equal.Right.Function)
            elif equal.Right.String is not None:
                self.Influenced_Equals[equal.Left.Function].add(equal.Right.String)
        if equal.Right.Function:
            if not self.Influenced_Equals.get(equal.Right.Function):
                self.Influenced_Equals[equal.Right.Function] = set()
            if equal.Left.Function:
                self.Influenced_Equals[equal.Right.Function].add(equal.Left.Function)
            elif equal.Left.String is not None:
                self.Influenced_Equals[equal.Right.Function].add(equal.Left.String)

    def process_condition(self, template, value):
        if isinstance(value, dict):
            if len(value) == 1:
                for func_name, func_value in value.items():
                    if func_name == cfnlint.helpers.FUNCTION_AND:
                        self.And = self.process_function(template, func_value)
                    elif func_name == cfnlint.helpers.FUNCTION_OR:
                        self.Or = self.process_function(template, func_value)
                    elif func_name == cfnlint.helpers.FUNCTION_NOT:
                        self.Not = self.process_function(template, func_value)
                    elif func_name == cfnlint.helpers.FUNCTION_EQUALS:
                        equal = Equals(func_value)
                        self.process_influenced_equal(equal)
                        self.Equals = equal
                    elif func_name == "Condition":
                        value = template.get("Conditions", {}).get(func_value, {})
                        try:
                            self.process_condition(template, value)
                        except ConditionParseError:
                            LOGGER.debug("Error parsing condition: %s", func_value)
            else:
                LOGGER.debug("Length of the object must be 1")
                raise ConditionParseError
        else:
            LOGGER.debug("Condition has to be an object")
            raise ConditionParseError

    def process_function(self, template, values):
        results = []
        for value in values:
            if isinstance(value, dict):
                if len(value) == 1:
                    for k, v in value.items():
                        if k == cfnlint.helpers.FUNCTION_EQUALS:
                            equal = Equals(v)
                            self.process_influenced_equal(equal)
                            results.append(equal)
                        elif k == "Condition":
                            condition = Condition(template, v)
                            results.append(condition)
                            for i_e_k, i_e_v in condition.Influenced_Equals.items():
                                if not self.Influenced_Equals.get(i_e_k):
                                    self.Influenced_Equals[i_e_k] = set()
                                for s_v in i_e_v:
                                    self.Influenced_Equals[i_e_k].add(s_v)
                        else:
                            condition = Condition(template, None, value)
                            results.append(condition)
                            for i_e_k, i_e_v in condition.Influenced_Equals.items():
                                if not self.Influenced_Equals.get(i_e_k):
                                    self.Influenced_Equals[i_e_k] = set()
                                for s_v in i_e_v:
                                    self.Influenced_Equals[i_e_k].add(s_v)

        return results


class Conditions:
    """All the conditions"""

    Conditions = None
    Equals = None
    Parameters = None

    def __init__(self, cfn):
        self.Conditions = {}
        self.Equals = {}
        self.Parameters = {}
        try:
            self.Equals = self._get_condition_equals(
                cfn.search_deep_keys(cfnlint.helpers.FUNCTION_EQUALS)
            )
            for condition_name in cfn.template.get("Conditions", {}):
                self.Conditions[condition_name] = Condition(
                    cfn.template, condition_name
                )
            # Configure parameters Allowed Values if they have them
            for parameter_name, parameter_values in cfn.template.get(
                "Parameters", {}
            ).items():
                # ALlowed Values must be a list so validate they are
                if isinstance(parameter_values.get("AllowedValues"), list):
                    # Any parameter in a condition could be used but would have to be done by
                    # Ref so build a ref to match for getting an equivalent hash
                    self.Parameters[
                        get_hash({"Ref": parameter_name})
                    ] = parameter_values.get("AllowedValues")
        except Exception as err:  # pylint: disable=W0703
            LOGGER.debug("While processing conditions got error: %s", err)

    def _get_condition_equals(self, equals):
        """
        Get Conditions based on Condition definitions
        Input: Cfn (Template)
        Output: List of hashes of Equal dict objects (Ref or FindInMap)
        """
        results = {}

        for equal in equals:
            if equal[0] == "Conditions":
                condition_name = equal[1]
                equals = equal[-1]
                if isinstance(equals, list):
                    if len(equals) == 2:
                        dict_hash_1 = None
                        value_1 = None
                        dict_hash_2 = None
                        value_2 = None
                        if isinstance(equals[0], dict):
                            dict_hash_1 = get_hash(equals[0])
                        elif isinstance(equals[0], str):
                            value_1 = equals[0]
                        elif isinstance(equals[0], int):
                            value_1 = str(equals[0])
                        if isinstance(equals[1], dict):
                            dict_hash_2 = get_hash(equals[1])
                        elif isinstance(equals[1], str):
                            value_2 = equals[1]
                        elif isinstance(equals[1], int):
                            value_2 = str(equals[1])

                        if dict_hash_1:
                            if dict_hash_1 not in results:
                                results[dict_hash_1] = []
                            if dict_hash_2:
                                results[dict_hash_1].append(
                                    {
                                        "Condition": condition_name,
                                        "Type": "dict",
                                        "Value": dict_hash_2,
                                    }
                                )
                            else:
                                results[dict_hash_1].append(
                                    {
                                        "Condition": condition_name,
                                        "Type": "string",
                                        "Value": value_2,
                                    }
                                )
                        if dict_hash_2:
                            if dict_hash_2 not in results:
                                results[dict_hash_2] = []
                            if dict_hash_1:
                                results[dict_hash_2].append(
                                    {
                                        "Condition": condition_name,
                                        "Type": "dict",
                                        "Value": dict_hash_1,
                                    }
                                )
                            else:
                                results[dict_hash_2].append(
                                    {
                                        "Condition": condition_name,
                                        "Type": "string",
                                        "Value": value_1,
                                    }
                                )
        return results

    def multiply_conditions(self, currents, condition, values):
        """Build out scenarios for when conditions don't match"""
        results = []
        if not currents:
            for value in values:
                new = {}
                new[condition] = value
                results.append(new)
        for current in currents:
            for value in values:
                new = copy(current)
                new[condition] = value
                results.append(new)

        return results

    def get_scenarios(self, conditions):
        """Get scenarios for all conditions provided"""
        matched_equals = {}
        matched_conditions = set()
        all_equals = {}
        results = []

        # When conditions don't properly get loaded (configuration error)
        # lets just return an empty list
        if not self.Conditions:
            return results

        for condition in conditions:
            # When one of the conditions don't exist we return an empty result
            if not self.Conditions.get(condition):
                return []

            for equal_key, equal_values in self.Conditions.get(
                condition
            ).Influenced_Equals.items():
                if not equal_key in all_equals:
                    all_equals[equal_key] = condition
                else:
                    if all_equals[equal_key] not in matched_conditions:
                        matched_conditions.add(all_equals[equal_key])
                    matched_conditions.add(condition)

        all_equals = {}
        for condition in matched_conditions:
            for equal_key, equal_values in self.Conditions.get(
                condition
            ).Influenced_Equals.items():
                if not equal_key in all_equals:
                    all_equals[equal_key] = set()
                else:
                    if not equal_key in matched_equals:
                        matched_equals[equal_key] = copy(all_equals[equal_key])
                for s_v in equal_values:
                    all_equals[equal_key].add(s_v)
                    if equal_key in matched_equals:
                        matched_equals[equal_key].add(s_v)

        diff = set(all_equals.keys()) - set(matched_equals.keys())
        if diff:
            diffKey = diff.pop()
            matched_equals[diffKey] = all_equals[diffKey]

        def multiply_equals(currents, s_hash, sets, parameter_values):
            """Multiply Equals when building scenarios"""
            results = []
            false_case = ""
            if not currents:
                # If the Parameter being REFed has Allowed Values use those instead
                if parameter_values:
                    for p_value in parameter_values:
                        # the allowed value must be an integer or string
                        # protecting against really badly formatted templates
                        if isinstance(p_value, (int, str)):
                            new = {}
                            # the allowed values could be numbers so force a string
                            new[s_hash] = str(p_value)
                            results.append(new)
                else:
                    for s_set in sets:
                        new = {}
                        new[s_hash] = s_set
                        false_case += s_set
                        results.append(new)
                    new = {}
                    new[s_hash] = false_case + ".bad"
                    results.append(new)
            for current in currents:
                # If the Parameter being REFed has Allowed Values use those instead
                if parameter_values:
                    for p_value in parameter_values:
                        # the allowed value must be an integer or string
                        # protecting against really badly formatted templates
                        if isinstance(p_value, (int, str)):
                            new = copy(current)
                            # the allowed values could be numbers so force a string
                            new[s_hash] = str(p_value)
                            results.append(new)
                else:
                    for s_set in sets:
                        new = copy(current)
                        new[s_hash] = s_set
                        false_case += s_set
                        results.append(new)
                    new = copy(current)
                    new[s_hash] = false_case + ".bad"
                    results.append(new)

            return results

        if not matched_conditions:
            # fail safe to not create a lot of unrelated scenarios. Just test if they are true/false
            # At this point this value is completely arbitrary and not configurable
            if len(conditions) > 4:
                LOGGER.info(
                    "Found %s conditions.  Limiting results to protect against heavy cpu time",
                    len(conditions),
                )
                true_results = []
                false_results = []
                for condition in conditions:
                    true_results = self.multiply_conditions(
                        true_results, condition, [True]
                    )
                    false_results = self.multiply_conditions(
                        false_results, condition, [False]
                    )
                results.extend(true_results)
                results.extend(false_results)
            else:
                for condition in conditions:
                    results = self.multiply_conditions(
                        results, condition, [True, False]
                    )

            return results

        if matched_conditions:
            scenarios = []
            for con_hash, sets in matched_equals.items():
                if len(scenarios) < 20:
                    scenarios = multiply_equals(
                        scenarios, con_hash, sets, self.Parameters.get(con_hash)
                    )

        for scenario in scenarios:
            r_condition = {}
            for condition in conditions:
                r_condition[condition] = self.Conditions.get(condition).test(scenario)

            if r_condition not in results:
                results.append(r_condition)

        return results



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import logging

import cfnlint.helpers
from cfnlint.rules import CloudFormationLintRule, RuleMatch

LOGGER = logging.getLogger("cfnlint")


class GetAtt(CloudFormationLintRule):
    """Check if GetAtt values are correct"""

    id = "E1010"
    shortdesc = "GetAtt validation of parameters"
    description = "Validates that GetAtt parameters are to valid resources and properties of those resources"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-getatt.html"
    tags = ["functions", "getatt"]

    def __init__(self):
        """Init"""
        super().__init__()
        self.propertytypes = []
        self.resourcetypes = []

    def initialize(self, cfn):
        resourcespecs = cfnlint.helpers.RESOURCE_SPECS[cfn.regions[0]]
        self.resourcetypes = resourcespecs["ResourceTypes"]
        self.propertytypes = resourcespecs["PropertyTypes"]

    def match(self, cfn):
        matches = []

        getatts = cfn.search_deep_keys("Fn::GetAtt")
        valid_getatts = cfn.get_valid_getatts()

        valid_attribute_functions = ["Ref"]

        for getatt in getatts:
            if len(getatt[-1]) < 2:
                message = "Invalid GetAtt for {0}"
                matches.append(
                    RuleMatch(getatt, message.format("/".join(map(str, getatt[:-1]))))
                )
                continue
            if isinstance(getatt[-1], str):
                resname, restype = getatt[-1].split(".", 1)
            else:
                resname = None
                restype = None
                if isinstance(getatt[-1][1], str):
                    resname = getatt[-1][0]
                    restype = ".".join(getatt[-1][1:])
                elif isinstance(getatt[-1][1], dict):
                    # You can ref the secondary part of a getatt

                    resname = getatt[-1][0]
                    restype = getatt[-1][1]
                    if len(restype) == 1:
                        for k in restype.keys():
                            if k not in valid_attribute_functions:
                                message = 'GetAtt only supports functions "{0}" for attributes at {1}'
                                matches.append(
                                    RuleMatch(
                                        getatt,
                                        message.format(
                                            ", ".join(
                                                map(str, valid_attribute_functions)
                                            ),
                                            "/".join(map(str, getatt[:-1])),
                                        ),
                                    )
                                )
                    else:
                        message = "Invalid GetAtt structure {0} at {1}"
                        matches.append(
                            RuleMatch(
                                getatt,
                                message.format(
                                    getatt[-1], "/".join(map(str, getatt[:-1]))
                                ),
                            )
                        )

                    # setting restype to None as we can't validate that anymore
                    restype = None
                else:
                    message = "Invalid GetAtt structure {0} at {1}"
                    matches.append(
                        RuleMatch(
                            getatt,
                            message.format(getatt[-1], "/".join(map(str, getatt[:-1]))),
                        )
                    )

            # only check resname if its set.  if it isn't it is because of bad structure
            # and an error is already provided
            if resname:
                if resname in valid_getatts:
                    if restype is not None:
                        # Check for maps
                        restypeparts = restype.split(".")
                        if (
                            restypeparts[0] in valid_getatts[resname]
                            and len(restypeparts) >= 2
                        ):
                            if (
                                valid_getatts[resname][restypeparts[0]].get("Type")
                                != "Map"
                            ):
                                message = "Invalid GetAtt {0}.{1} for resource {2}"
                                matches.append(
                                    RuleMatch(
                                        getatt[:-1],
                                        message.format(resname, restype, getatt[1]),
                                    )
                                )
                            else:
                                continue
                        if (
                            restype not in valid_getatts[resname]
                            and "*" not in valid_getatts[resname]
                        ):
                            message = "Invalid GetAtt {0}.{1} for resource {2}"
                            matches.append(
                                RuleMatch(
                                    getatt[:-1],
                                    message.format(resname, restype, getatt[1]),
                                )
                            )
                else:
                    modules = cfn.get_modules()
                    if any(resname.startswith(s) for s in modules.keys()):
                        LOGGER.debug(
                            "Will consider valid getatt %s because it references a resource within a module",
                            resname,
                        )
                    else:
                        message = "Invalid GetAtt {0}.{1} for resource {2}"
                        matches.append(
                            RuleMatch(
                                getatt, message.format(resname, restype, getatt[1])
                            )
                        )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from datetime import datetime

from cfnlint.rules import RuleMatch
from cfnlint.rules.resources.lmbd.DeprecatedRuntime import DeprecatedRuntime


class DeprecatedRuntimeEnd(DeprecatedRuntime):
    """Check if EOL Lambda Function Runtimes are used"""

    id = "E2531"
    shortdesc = "Check if EOL Lambda Function Runtimes are used"
    description = (
        "Check if an EOL Lambda Runtime is specified and give an error if used. "
    )
    source_url = (
        "https://docs.aws.amazon.com/lambda/latest/dg/runtime-support-policy.html"
    )
    tags = ["resources", "lambda", "runtime"]

    def check_runtime(self, runtime_value, path):
        """Check if the given runtime is valid"""
        matches = []

        runtime = self.deprecated_runtimes.get(runtime_value)
        if runtime:
            if datetime.strptime(runtime["deprecated"], "%Y-%m-%d") < self.current_date:
                message = "Deprecated runtime ({0}) specified. Updating disabled since {1}. Please consider updating to {2}"
                matches.append(
                    RuleMatch(
                        path,
                        message.format(
                            runtime_value, runtime["deprecated"], runtime["successor"]
                        ),
                    )
                )
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import logging

import cfnlint.conditions
import cfnlint.helpers
import cfnlint.rules
from cfnlint.decode.node import TemplateAttributeError, dict_node
from cfnlint.decorators.refactored import refactored
from cfnlint.graph import Graph
from cfnlint.helpers import PSEUDOPARAMS
from cfnlint.rules import CloudFormationLintRule as _CloudFormationLintRule
from cfnlint.rules import Match as _Match
from cfnlint.rules import ParseError as _ParseError
from cfnlint.rules import RuleError as _RuleError
from cfnlint.rules import RuleMatch as _RuleMatch
from cfnlint.rules import RulesCollection as _RulesCollection
from cfnlint.rules import TransformError as _TransformError
from cfnlint.runner import Runner as _Runner
from cfnlint.template import Template as _Template
from cfnlint.transform import Transform

LOGGER = logging.getLogger(__name__)


@refactored(
    "RuleMatch is refactored and deprecated. Please use cfnlint.rules.RuleMatch"
)
class RuleMatch(_RuleMatch):
    """Refactored class RuleMatch"""


@refactored("Match is refactored and deprecated. Please use cfnlint.rules.Match")
class Match(_Match):
    """Refactored class Match"""


@refactored(
    "RulesCollection is refactored and deprecated. Please use cfnlint.rules.RulesCollection"
)
class RulesCollection(_RulesCollection):
    """Refactored class Rules Collection"""


@refactored(
    "CloudFormationLintRule is refactored and deprecated. Please use cfnlint.rules.CloudFormationLintRule"
)
class CloudFormationLintRule(_CloudFormationLintRule):
    """Refactored class Rules CloudFormationLintRule"""


@refactored(
    "ParseError is refactored and deprecated. Please use cfnlint.rules.ParseError"
)
class ParseError(_ParseError):
    """Refactored class Rules ParseError"""


@refactored(
    "TransformError is refactored and deprecated. Please use cfnlint.rules.TransformError"
)
class TransformError(_TransformError):
    """Refactored class Rules TransformError"""


@refactored(
    "RuleError is refactored and deprecated. Please use cfnlint.rules.RuleError"
)
class RuleError(_RuleError):
    """Refactored class Rules RuleError"""


@refactored(
    "Template is refactored and deprecated. Please use Template in cfnlint.template"
)
class Template(_Template):
    """Refactored class Template"""


@refactored("Runner is refactored and deprecated. Please use Runner in cfnlint.runner")
class Runner(_Runner):
    """Refactored class Runner"""
